"""Parsing of UML model elements from string tests."""

import pytest

from gaphor import UML
from gaphor.core.eventmanager import EventManager
from gaphor.core.format import parse
from gaphor.core.modeling import ElementFactory
from gaphor.UML.umllex import attribute_pat, operation_pat, parameters_pat


def dump_prop(prop):
    attribute_pat.match(prop)


def dump_oper(oper):
    if m := operation_pat.match(oper):
        g = m.group
    else:
        # set name to oper
        return
    if g("params"):
        params = g("params")
        while params:
            m = parameters_pat.match(params)
            g = m.group
            params = g("rest")


dump_prop("#/name")
dump_prop('+ / name : str[1..*] = "aap" { static }')
dump_prop('+ / name : str[*] = "aap" { static }')

dump_oper('myfunc(aap:str = "aap", out two): type')
dump_oper("   myfunc2 ( ): type")
dump_oper('myfunc(aap:str[1] = "aap" { tag1, tag2 }, out two {tag3}): type')


@pytest.fixture
def factory():
    event_manager = EventManager()
    return ElementFactory(event_manager)


def test_parse_property_simple(factory):
    """Test simple property parsing."""
    a = factory.create(UML.Property)
    parse(a, "myattr")
    assert not a.isDerived
    assert "myattr" == a.name
    assert a.typeValue is None
    assert UML.recipes.get_multiplicity_lower_value(a) is None, a.lowerValue
    assert UML.recipes.get_multiplicity_upper_value(a) is None, a.upperValue
    assert a.defaultValue is None


def test_parse_property_complex_1(factory):
    """Test complex property parsing."""
    a = factory.create(UML.Property)

    parse(a, '+ / name : str[0..*] = "aap" { static }# and a note')
    assert "public" == a.visibility
    assert a.isDerived
    assert "name" == a.name
    assert "str" == a.typeValue
    assert UML.recipes.get_multiplicity_lower_value(a) == 0
    assert UML.recipes.get_multiplicity_upper_value(a) == "*"
    assert UML.recipes.get_default_value_as_string(a) == '"aap"'
    assert "and a note" == a.note


def test_parse_property_complex_2(factory):
    """Test complex property parsing."""
    a = factory.create(UML.Property)

    parse(a, '+ / name : str[0..*] = "aap bbq" { static }# and a note')
    assert "public" == a.visibility
    assert a.isDerived
    assert "name" == a.name
    assert "str" == a.typeValue
    assert UML.recipes.get_multiplicity_lower_value(a) == 0
    assert UML.recipes.get_multiplicity_upper_value(a) == "*"
    assert UML.recipes.get_default_value_as_string(a) == '"aap bbq"'
    assert "and a note" == a.note


def test_parse_property_with_space_in_name(factory):
    a = factory.create(UML.Property)

    parse(a, "+ name with space : str")

    assert "public" == a.visibility
    assert "name with space" == a.name
    assert "str" == a.typeValue


def test_parse_property_with_default_value_and_note(factory):
    a = factory.create(UML.Property)

    parse(a, "name=3 #note")

    assert "name" == a.name
    assert "3" == UML.recipes.get_default_value_as_string(a)
    assert "note" == a.note


def test_parse_property_with_square_brackets(factory):
    a = factory.create(UML.Property)
    parse(a, "attr[]")

    assert "attr" == a.name
    assert UML.recipes.get_multiplicity_lower_value(a) is None
    assert UML.recipes.get_multiplicity_upper_value(a) == "*"


@pytest.mark.parametrize(
    "text,type_value",
    [
        ("attr: int | str", "int | str"),
        ("attr: int | str | bool | other", "int | str | bool | other"),
        ("attr: int|str|bool", "int|str|bool"),
        ("attr: my int|with space|some bool", "my int|with space|some bool"),
        ("attr: Generic type<Type>", "Generic type<Type>"),
        ("attr: Generic<Type> with extras", "Generic<Type> with extras"),
        ("attr: Generic<Type|Other>", "Generic<Type|Other>"),
    ],
)
def test_parse_property_with_union_type(factory, text, type_value):
    a = factory.create(UML.Property)
    parse(a, text)

    assert "attr" == a.name
    assert type_value == a.typeValue


def test_parse_property_invalid(factory):
    """Test parsing property with invalid syntax."""
    a = factory.create(UML.Property)

    parse(a, '+ name = str[*] = "aap" { static }')
    assert '+ name = str[*] = "aap" { static }' == a.name
    assert not a.isDerived
    assert not a.typeValue
    assert not a.lowerValue
    assert not a.upperValue
    assert not a.defaultValue


def test_parse_association_end(factory):
    """Test parsing of association end."""
    a = factory.create(UML.Association)
    p = factory.create(UML.Property)
    p.association = a

    parse(p, "end")
    assert "end" == p.name
    assert not p.typeValue
    assert not p.lowerValue
    assert not p.upperValue
    assert not p.defaultValue


def test_parse_association_end_multiplicity(factory):
    """Test parsing of multiplicity."""
    a = factory.create(UML.Association)
    p = factory.create(UML.Property)
    p.association = a
    parse(p, "0..2 { tag }")
    assert p.name is None
    assert not p.typeValue
    assert UML.recipes.get_multiplicity_lower_value(p) == 0
    assert UML.recipes.get_multiplicity_upper_value(p) == 2
    assert not p.defaultValue


def test_parse_association_end_multiplicity2(factory):
    """Test parsing of multiplicity with multiline constraints."""
    a = factory.create(UML.Association)
    p = factory.create(UML.Property)
    p.association = a
    parse(p, "0..2 { tag1, \ntag2}")
    assert p.name is None
    assert not p.typeValue
    assert UML.recipes.get_multiplicity_lower_value(p) == 0
    assert UML.recipes.get_multiplicity_upper_value(p) == 2
    assert not p.defaultValue


@pytest.mark.parametrize("text", ["-0..2", "-[0..2]"])
def test_parse_association_end_visibility_multiplicity_no_name(text, factory):
    """Test parsing of multiplicity with multiline constraints."""
    a = factory.create(UML.Association)
    p = factory.create(UML.Property)
    p.association = a
    parse(p, text)
    assert not p.name
    assert not p.typeValue
    assert p.visibility == "private"
    assert UML.recipes.get_multiplicity_lower_value(p) == 0
    assert UML.recipes.get_multiplicity_upper_value(p) == 2
    assert not p.defaultValue


def test_parse_association_end_derived_end(factory):
    """Test parsing derived association end."""
    a = factory.create(UML.Association)
    p = factory.create(UML.Property)
    p.association = a
    parse(p, "-/end name[*] { mytag}")
    assert "private" == p.visibility
    assert p.isDerived
    assert "end name" == p.name
    assert not p.typeValue
    assert UML.recipes.get_multiplicity_lower_value(p) is None
    assert UML.recipes.get_multiplicity_upper_value(p) == "*"
    assert not p.defaultValue


def test_parse_association_end_derived_end_without_name(factory):
    """Test parsing derived association end without name."""
    a = factory.create(UML.Association)
    p = factory.create(UML.Property)
    p.association = a
    parse(p, "-/*")
    assert "private" == p.visibility
    assert p.isDerived
    assert not p.name
    assert not p.typeValue
    assert UML.recipes.get_multiplicity_lower_value(p) is None
    assert UML.recipes.get_multiplicity_upper_value(p) == "*"
    assert not p.defaultValue


def test_parse_association_end_with_type(factory):
    """Test parsing of association end, type is ignored."""
    a = factory.create(UML.Association)
    p = factory.create(UML.Property)
    p.association = a

    parse(p, "end: TypeVal")
    assert "end" == p.name
    assert not p.typeValue
    assert not p.lowerValue
    assert not p.upperValue
    assert not p.defaultValue


def test_parse_association_end_with_note(factory):
    a = factory.create(UML.Association)
    p = factory.create(UML.Property)
    p.association = a

    parse(p, "end # some note")
    assert "end" == p.name
    assert "some note" == p.note


def test_parse_association_end_with_square_brackets(factory):
    a = factory.create(UML.Association)
    p = factory.create(UML.Property)
    p.association = a

    parse(p, "end[]")
    assert "end" == p.name
    assert UML.recipes.get_multiplicity_lower_value(p) is None
    assert UML.recipes.get_multiplicity_upper_value(p) == "*"


def test_parse_operation(factory):
    """Test parsing simple operation."""
    o = factory.create(UML.Operation)
    parse(o, "myfunc()")
    assert "myfunc" == o.name
    assert not o.ownedParameter


def test_parse_operation_return(factory):
    """Test parsing operation with return value."""
    o = factory.create(UML.Operation)
    parse(o, "+ myfunc(): int")
    assert "myfunc" == o.name
    assert "int" == o.ownedParameter[0].typeValue
    assert "public" == o.visibility


def test_parse_operation_2_params(factory):
    """Test parsing of operation with two parameters."""
    o = factory.create(UML.Operation)
    parse(o, "# myfunc2 (a: str, b: int = 3 {  static}): float")
    assert "myfunc2" == o.name
    assert "protected" == o.visibility
    assert len(o.ownedParameter) == 3
    assert "float" == o.ownedParameter[0].typeValue
    assert "a" == o.ownedParameter[1].name
    assert "str" == o.ownedParameter[1].typeValue
    assert UML.recipes.get_default_value_as_string(o.ownedParameter[1]) is None
    assert "b" == o.ownedParameter[2].name
    assert "int" == o.ownedParameter[2].typeValue
    assert UML.recipes.get_default_value_as_string(o.ownedParameter[2]) == "3"


def test_parse_operation_1_param(factory):
    """Test parsing of operation with one parameter."""
    o = factory.create(UML.Operation)
    parse(o, "- myfunc2 (a: node): double")
    assert "myfunc2" == o.name
    assert "private" == o.visibility
    assert len(o.ownedParameter) == 2
    assert "double" == o.ownedParameter[0].typeValue
    assert "a" == o.ownedParameter[1].name
    assert "node" == o.ownedParameter[1].typeValue
    assert o.ownedParameter[1].defaultValue is None


def test_parse_operation_with_spaces(factory):
    o = factory.create(UML.Operation)
    parse(o, "- name with space (param with space: some node ): return value")
    assert "name with space" == o.name
    assert "return value" == o.ownedParameter[0].typeValue
    assert "param with space" == o.ownedParameter[1].name
    assert "some node" == o.ownedParameter[1].typeValue


def test_parse_operation_with_union_types(factory):
    o = factory.create(UML.Operation)
    parse(o, "- oper(param with space: int | str ): bool | None")
    assert "oper" == o.name
    assert "bool | None" == o.ownedParameter[0].typeValue
    assert "int | str" == o.ownedParameter[1].typeValue
    assert "param with space" == o.ownedParameter[1].name


def test_parse_operation_invalid_syntax(factory):
    """Test operation parsing with invalid syntax."""
    o = factory.create(UML.Operation)
    parse(o, "- myfunc2: myType2")
    assert "- myfunc2: myType2" == o.name


def test_parse_operation_with_note(factory):
    """Test parsing simple operation."""
    o = factory.create(UML.Operation)
    parse(o, "myfunc() # and a note")
    assert "myfunc" == o.name
    assert "and a note" == o.note


def test_parse_operation_with_square_brackets(factory):
    o: UML.Operation = factory.create(UML.Operation)
    parse(o, "myfunc(args: string[])")
    p = o.ownedParameter[0]
    assert "args" == p.name
    assert "string" == p.typeValue
    assert UML.recipes.get_multiplicity_lower_value(p) is None
    assert UML.recipes.get_multiplicity_upper_value(p) == "*"


def test_parse_operation_return_with_square_brackets(factory):
    o: UML.Operation = factory.create(UML.Operation)
    parse(o, "myfunc(): string[]")
    p = o.ownedParameter[0]
    assert "return" == p.direction
    assert "string" == p.typeValue
    assert UML.recipes.get_multiplicity_lower_value(p) is None
    assert UML.recipes.get_multiplicity_upper_value(p) == "*"


@pytest.mark.parametrize(
    "operation,type_value",
    [
        ["+ test(param: Generic<Type>)", "Generic<Type>"],
        ["+ test(param: Generic with space <Type>)", "Generic with space <Type>"],
        [
            "+ test(param: Generic with space <Type> | Other<Type|Other>)",
            "Generic with space <Type> | Other<Type|Other>",
        ],
        [
            "+ test(param: Generic <Type> | Other<Type> | Third<Type>)",
            "Generic <Type> | Other<Type> | Third<Type>",
        ],
    ],
)
def test_parse_operation_with_template_attribute(operation, type_value, factory):
    o: UML.Operation = factory.create(UML.Operation)
    parse(o, operation)
    p = o.ownedParameter[0]
    assert "param" == p.name
    assert type_value == p.typeValue



"""Test classes."""

from gaphor import UML
from gaphor.UML.classes.interface import Folded, InterfaceItem


def test_interface_creation(create):
    """Test interface creation."""
    iface = create(InterfaceItem, UML.Interface)
    assert isinstance(iface.subject, UML.Interface)


def test_folded_interface_persistence(create, loader, saver, element_factory):
    """Test folded interface saving/loading."""
    iface = create(InterfaceItem, UML.Interface)

    # note: assembly folded mode..
    iface.folded = Folded.REQUIRED

    data = saver()
    loader(data)

    interfaces = list(element_factory.select(InterfaceItem))
    assert len(interfaces) == 1
    # ... gives provided folded mode on load;
    # correct folded mode is determined by connections, which will be
    # recreated later, i.e. required folded mode will be set when
    # implementation connects to the interface and Folded.PROVIDED
    # is equal to interfaces[0].folded



from __future__ import annotations

import pytest
from gi.repository import Gio, GObject

from gaphor import UML
from gaphor.abc import ModelingLanguage
from gaphor.core.modeling import Base, Diagram, ModelReady
from gaphor.i18n import gettext
from gaphor.services.componentregistry import ComponentRegistry
from gaphor.services.modelinglanguage import ModelingLanguageService
from gaphor.ui.modelbrowser import (
    ElementDragData,
    ModelBrowser,
    SearchEngine,
    get_first_selected_item,
    list_item_drop_drop,
)
from gaphor.UML import recipes
from gaphor.UML.treemodel import TreeModel as UMLTreeModel


@pytest.fixture
def component_registry(event_manager, element_factory):
    component_registry = ComponentRegistry()
    component_registry.register("event_manager", event_manager)
    component_registry.register("element_factory", element_factory)
    yield component_registry
    component_registry.shutdown()


@pytest.fixture
def modeling_language(event_manager, properties):
    return DummyModelingLanguageService(event_manager, properties)


@pytest.fixture
def model_browser(
    event_manager, component_registry, element_factory, modeling_language
):
    model_browser = ModelBrowser(
        event_manager, component_registry, element_factory, modeling_language
    )
    model_browser.open()
    yield model_browser
    model_browser.close()


class ItemChangedHandler:
    def __init__(self):
        self.added = 0
        self.removed = 0
        self.positions = []

    def __call__(self, _obj, position, removed, added):
        self.positions.append(position)
        self.removed += removed
        self.added += added


def test_show_item_in_tree_list_model(model_browser, element_factory):
    class_ = element_factory.create(UML.Class)
    package = element_factory.create(UML.Package)

    class_.package = package

    pos = model_browser.select_element(class_)
    tree_model = model_browser.selection

    assert pos == 1
    assert tree_model.get_item(0).get_item().element is package
    assert tree_model.get_item(1).get_item().element is class_


def test_model_browser_remove_element(model_browser, element_factory):
    tree_model = model_browser.model.root
    element = element_factory.create(UML.Class)
    items_changed = ItemChangedHandler()
    tree_model.connect("items-changed", items_changed)

    element.unlink()

    assert len(tree_model) == 0
    assert items_changed.removed == 1


def test_model_browser_add_nested_element(model_browser, element_factory):
    tree_model = model_browser.model
    class_ = element_factory.create(UML.Class)
    package = element_factory.create(UML.Package)

    class_.package = package
    package_item = tree_model.tree_item_for_element(package)
    child_model = tree_model.child_model(package_item)

    assert package_item in tree_model.root
    assert tree_model.tree_item_for_element(class_) in child_model


def test_model_browser_unset_nested_element(model_browser, element_factory):
    tree_model = model_browser.model
    class_ = element_factory.create(UML.Class)
    package = element_factory.create(UML.Package)

    class_.package = package
    del class_.owningPackage
    package_item = tree_model.tree_item_for_element(package)
    class_item = tree_model.tree_item_for_element(class_)
    child_model = tree_model.child_model(package_item)

    assert package_item is not None
    assert class_item is not None
    assert child_model is None


def test_model_browser_remove_nested_element(model_browser, element_factory):
    tree_model = model_browser.model
    class_ = element_factory.create(UML.Class)
    package = element_factory.create(UML.Package)

    class_.package = package
    package_item = tree_model.tree_item_for_element(package)
    assert tree_model.child_model(package_item)

    class_.unlink()

    child_model = tree_model.child_model(package_item)
    assert tree_model.tree_item_for_element(package) is not None
    assert child_model is None


def test_change_from_member_to_owner(model_browser, element_factory):
    tree_model = model_browser.model
    property = element_factory.create(UML.Property)
    association = element_factory.create(UML.Association)
    klass = element_factory.create(UML.Class)

    association.memberEnd = property
    klass.ownedAttribute = property

    association_item = tree_model.tree_item_for_element(association)
    klass_item = tree_model.tree_item_for_element(klass)
    property_item = tree_model.tree_item_for_element(property)
    association_model = tree_model.branches.get(association_item)
    klass_model = tree_model.branches.get(klass_item)

    assert not association_model
    assert property_item in klass_model.elements


def test_element_name_changed(model_browser, element_factory):
    class_ = element_factory.create(UML.Class)
    tree_item = model_browser.model.tree_item_for_element(class_)

    class_.name = "foo"

    weight, style = tree_item.attributes.get_attributes()

    assert tree_item.readonly_text == "foo"
    assert weight.as_int().value == 400
    assert style.as_int().value == 0


def test_element_weight_changed(model_browser, element_factory):
    diagram = element_factory.create(UML.Diagram)
    tree_item = model_browser.model.tree_item_for_element(diagram)
    weight, style = tree_item.attributes.get_attributes()

    assert weight.as_int().value == 700
    assert style.as_int().value == 0


def test_element_style_changed(model_browser, element_factory):
    class_ = element_factory.create(UML.Class)
    tree_item = model_browser.model.tree_item_for_element(class_)

    class_.isAbstract = True
    weight, style = tree_item.attributes.get_attributes()

    assert weight.as_int().value == 400
    assert style.as_int().value == 2


def test_model_browser_model_ready(
    event_manager, component_registry, element_factory, modeling_language
):
    class_ = element_factory.create(UML.Class)
    package = element_factory.create(UML.Package)
    class_.package = package
    model_browser = ModelBrowser(
        event_manager, component_registry, element_factory, modeling_language
    )
    model_browser.open()
    tree_model = model_browser.model

    event_manager.handle(ModelReady(element_factory))

    assert tree_model.tree_item_for_element(package) is not None
    assert tree_model.tree_item_for_element(class_) is None

    model_browser.close()


def test_default_model_browser_model_is_UML(model_browser):
    assert type(model_browser.model) is UMLTreeModel


def test_model_browser_language_changed(model_browser, modeling_language):
    modeling_language.select_modeling_language("Dummy")

    assert type(model_browser.model) is DummyTreeModel


def test_model_browser_should_not_change_for_similar_model_types(
    model_browser, modeling_language
):
    current_model = model_browser.model

    modeling_language.select_modeling_language("SysML")

    assert model_browser.model is current_model


def test_tree_model_expand_to_relationship(model_browser, element_factory):
    association = element_factory.create(UML.Association)
    package = element_factory.create(UML.Package)
    association.package = package

    pos = model_browser.select_element(association)

    assert pos == 2
    assert model_browser.get_selected_element() is association


def test_create_diagram(model_browser, element_factory):
    model_browser.tree_view_create_diagram("cls")

    diagram = next(element_factory.select(Diagram))

    assert diagram.diagramType == "cls"


def test_delete_element(model_browser, element_factory):
    klass = element_factory.create(UML.Class)
    model_browser.select_element(klass)
    assert model_browser.get_selected_element() is klass

    model_browser.tree_view_delete()

    assert not element_factory.lselect()


def test_association_end(model_browser: ModelBrowser, element_factory):
    tree_model = model_browser.model
    head_class = element_factory.create(UML.Class)
    tail_class = element_factory.create(UML.Class)
    association = recipes.create_association(head_class, tail_class)

    recipes.set_navigability(association, association.memberEnd[0], True)
    recipes.set_navigability(association, association.memberEnd[0], None)

    association_item = tree_model.tree_item_for_element(association)
    association_children = tree_model.child_model(association_item)

    assert association.memberEnd[0] in (t.element for t in association_children)
    assert association.memberEnd[1] in (t.element for t in association_children)


def test_navigable_association_end(model_browser: ModelBrowser, element_factory):
    tree_model = model_browser.model
    head_class = element_factory.create(UML.Class)
    tail_class = element_factory.create(UML.Class)
    association = recipes.create_association(head_class, tail_class)

    recipes.set_navigability(association, association.memberEnd[0], True)

    association_item = tree_model.tree_item_for_element(association)
    tail_class_item = tree_model.tree_item_for_element(tail_class)
    association_children = tree_model.child_model(association_item)
    tail_class_children = tree_model.child_model(tail_class_item)

    assert association.memberEnd[0] not in (t.element for t in association_children)
    assert association.memberEnd[0] in (t.element for t in tail_class_children)


def test_search_next(model_browser, element_factory):
    class_a = element_factory.create(UML.Class)
    class_a.name = "a"
    class_b = element_factory.create(UML.Class)
    class_b.name = "b"

    search_engine = SearchEngine(model_browser.model, model_browser.tree_view)
    model_browser.select_element(class_a)
    assert model_browser.get_selected_element() is class_a

    search_engine.search_next("b")

    assert model_browser.get_selected_element() is class_b


def test_search_text_changed(model_browser, element_factory):
    class_a = element_factory.create(UML.Class)
    class_a.name = "a"
    class_b = element_factory.create(UML.Class)
    class_b.name = "b"

    search_engine = SearchEngine(model_browser.model, model_browser.tree_view)
    model_browser.select_element(class_a)
    assert model_browser.get_selected_element() is class_a

    search_engine.text_changed("b")

    assert model_browser.get_selected_element() is class_b


def test_generalization_text(model_browser, element_factory):
    general = element_factory.create(UML.Class)
    general.name = "General"
    specific = element_factory.create(UML.Class)
    specific.name = "Specific"
    generalization = element_factory.create(UML.Generalization)
    generalization.specific = specific
    generalization.general = general

    model = model_browser.model
    tree_item = model.tree_item_for_element(specific)
    branch = model.branches[tree_item]
    assert tree_item
    assert branch.relationships[0].element is generalization
    assert branch.relationships[0].readonly_text == gettext("general: General")


def test_drop_multiple_elements(model_browser, element_factory, event_manager):
    class_a = element_factory.create(UML.Class)
    class_a.name = "a"
    class_b = element_factory.create(UML.Class)
    class_b.name = "b"
    gen = element_factory.create(UML.Generalization)
    gen.general = class_a
    gen.specific = class_b
    package = element_factory.create(UML.Package)

    model_browser.select_element(package)
    list_item = MockRowItem(get_first_selected_item(model_browser.selection))

    target = MockDropTarget()
    drag_data = ElementDragData(elements=[class_a, class_b])
    list_item_drop_drop(target, drag_data, 0, 10, list_item, event_manager)

    assert class_b.owner is package
    assert class_a.owner is package


def test_unlink_element_should_not_collapse_branch(
    model_browser, element_factory, event_manager
):
    package = element_factory.create(UML.Package)
    package.name = "p"
    class_a = element_factory.create(UML.Class)
    class_a.package = package
    class_a.name = "a"
    class_b = element_factory.create(UML.Class)
    class_b.package = package
    class_b.name = "b"

    row0 = model_browser.selection.get_item(0)
    row0.set_expanded(True)

    class_a.unlink()

    assert row0.get_item().element is package
    assert row0.get_expanded()
    assert model_browser.selection.get_item(1).get_item().element is class_b


def test_multiplicity_element_should_not_end_up_in_root(model_browser, element_factory):
    port = element_factory.create(UML.Port)
    prop = element_factory.create(UML.Property)
    connector = UML.recipes.create_connector(port, prop)

    model = model_browser.model

    assert model.tree_item_for_element(connector.end[0]) is None


def test_stereotype_base_class_should_not_end_up_in_root(
    model_browser, element_factory
):
    profile = element_factory.create(UML.Profile)
    profile.name = "profile"
    metaclass = element_factory.create(UML.Class)
    metaclass.package = profile
    metaclass.name = "Metaclass"
    stereotype = element_factory.create(UML.Stereotype)
    stereotype.package = profile
    stereotype.name = "Stereotype"

    relation = recipes.create_extension(metaclass, stereotype)
    relation.package = profile

    row0 = model_browser.selection.get_item(0)
    assert row0.get_item().element is profile  # not property


def test_drag_and_drop_parent_on_child(model_browser, element_factory, event_manager):
    root = element_factory.create(UML.Package)
    parent = element_factory.create(UML.Package)
    parent.nestingPackage = root
    parent_diagram = element_factory.create(UML.Diagram)
    parent_diagram.element = parent
    child = element_factory.create(UML.Package)
    child.nestingPackage = parent
    child_diagram = element_factory.create(UML.Diagram)
    child_diagram.element = child

    model_browser.select_element(parent)
    list_item = MockRowItem(get_first_selected_item(model_browser.selection))

    target = MockDropTarget()
    drag_data = ElementDragData(elements=[parent])
    list_item_drop_drop(target, drag_data, 0, 10, list_item, event_manager)

    assert model_browser.select_element(child) is not None
    assert child.owner is parent
    assert parent.owner is root


class MockRowItem:
    def __init__(self, list_item):
        self.list_item = list_item

    def get_item(self):
        return self.list_item


class MockDropTarget:
    def get_widget(self):
        class Widget:
            def remove_css_class(self, _):
                pass

        return Widget()


class DummyModelingLanguageService(ModelingLanguageService):
    def __init__(self, event_manager=None, properties=None):
        super().__init__(event_manager, properties)
        self._modeling_languages["Dummy"] = DummyModelingLanguage()


class DummyModelingLanguage(ModelingLanguage):
    @property
    def name(self) -> str:
        """Human-readable name of the modeling language."""
        return "Dummy"

    @property
    def toolbox_definition(self):
        raise ValueError("toolbox_definition not implemented")

    @property
    def diagram_types(self):
        return ()

    @property
    def element_types(self):
        return ()

    @property
    def model_browser_model(self):
        return DummyTreeModel

    def lookup_element(self, name: str, ns: str | None = None):
        raise ValueError("toolbox_definition not implemented")


class DummyTreeModel:
    def __init__(self, on_select=None, on_sync=None):
        self._root = Gio.ListStore.new(DummyTreeItem.__gtype__)

    @property
    def root(self) -> Gio.ListStore:
        return self._root

    def child_model(self, item: DummyTreeItem) -> Gio.ListStore | None:
        return None

    @property
    def template(self) -> str:
        return ""

    def tree_item_sort(self, a, b) -> int:
        return 0

    def should_expand(self, item: DummyTreeItem, element: Base) -> bool:
        return True

    def shutdown(self) -> None:
        pass


class DummyTreeItem(GObject.Object):
    def __init__(self, element: Base | None):
        super().__init__()
        self.element = element



import pytest

from gaphor.UML.actions.activitynodes import ForkNodeItem


class Saver:
    def __call__(self, name, val):
        setattr(self, name, val)


@pytest.fixture
def saver():
    return Saver()


def test_forknode_save_default_height(diagram, saver):
    fork: ForkNodeItem = diagram.create(ForkNodeItem)
    diagram.update({fork})

    fork.save(saver)

    assert saver.height == 30
    assert saver.matrix[5] == 0


def test_forknode_save_height_0(diagram, saver):
    fork: ForkNodeItem = diagram.create(ForkNodeItem)

    fork.handles()[0].pos.y = -100
    fork.handles()[1].pos.y = 0
    diagram.update({fork})

    fork.save(saver)

    assert saver.height == 100
    assert saver.matrix[5] == -100


def test_forknode_save_height_1(diagram, saver):
    fork: ForkNodeItem = diagram.create(ForkNodeItem)

    fork.handles()[0].pos.y = 0
    fork.handles()[1].pos.y = 100
    diagram.update({fork})

    fork.save(saver)

    assert saver.height == 100
    assert saver.matrix[5] == 0



from gaphor import UML
from gaphor.core import Transaction
from gaphor.core.modeling import Diagram
from gaphor.UML.classes import ClassItem


class ViewMock:
    def __init__(self, diagram):
        self.updates = []
        diagram.register_view(self)

    def request_update(self, items=(), removed_items=()):
        self.updates.append((items, removed_items))


def test_undo_should_remove_shown_item_on_diagram(
    event_manager, element_factory, undo_manager
):
    with Transaction(event_manager):
        diagram = element_factory.create(Diagram)

    view = ViewMock(diagram)

    with Transaction(event_manager):
        cls = diagram.create(ClassItem, subject=element_factory.create(UML.Class))

    undo_manager.undo_transaction()

    _items, removed_items = view.updates[-1]

    assert cls in removed_items, view.updates


def test_redo_should_show_item_on_diagram(event_manager, element_factory, undo_manager):
    with Transaction(event_manager):
        diagram = element_factory.create(Diagram)

    view = ViewMock(diagram)

    with Transaction(event_manager):
        cls = diagram.create(ClassItem, subject=element_factory.create(UML.Class))

    undo_manager.undo_transaction()
    undo_manager.redo_transaction()

    items, _removed_items = view.updates[-1]

    assert cls.id in (item.id for item in items), items



import os
from pathlib import Path

import pytest
from gi.repository import Gio, GLib, Gtk

from gaphor.ui.filedialog import open_file_dialog, pretty_path, save_file_dialog


class TaskMock:
    def __init__(self, error=False):
        self._error = error

    def had_error(self):
        return self._error


class FileDialogMock(Gtk.FileDialog):
    def __init__(self):
        super().__init__()
        self._save_response = Gio.File.parse_name("/unset")
        self._error = False
        self.calls = {
            "save": 0,
            "open": 0,
            "open_multiple": 0,
        }

    async def save(self, parent):
        self.calls["save"] += 1
        if self._error:
            raise GLib.Error(message="", domain=Gtk.dialog_error_quark(), code=2)
        return self._save_response

    async def open(self, parent):
        self.calls["open"] += 1
        if self._error:
            raise GLib.Error(message="", domain=Gtk.dialog_error_quark(), code=2)
        return self._save_response

    async def open_multiple(self, parent):
        self.calls["open_multiple"] += 1
        if self._error:
            raise GLib.Error(message="", domain=Gtk.dialog_error_quark(), code=2)
        return self._save_response

    def define_response(self, response):
        if isinstance(response, list):
            self._save_response = Gio.ListStore(item_type=Gio.File)
            for r in response:
                self._save_response.append(Gio.File.parse_name(r))
        else:
            self._save_response = Gio.File.parse_name(response)

    def define_error(self, error):
        self._error = error


@pytest.fixture
def file_dialog(monkeypatch):
    stub = FileDialogMock()

    def new_file_dialog():
        return stub

    monkeypatch.setattr("gi.repository.Gtk.FileDialog.new", new_file_dialog)
    return stub


@pytest.mark.asyncio
async def test_save_dialog_cancelled(file_dialog):
    file_dialog.define_response("/test/path/model.gaphor")
    file_dialog.define_error(True)
    selected_file = await save_file_dialog(
        "title",
        Path("testfile.gaphor"),
        filters=[],
    )

    assert file_dialog.calls["save"] == 1
    assert selected_file is None


@pytest.mark.asyncio
async def test_save_dialog(file_dialog):
    file_dialog.define_response("/test/path/model.gaphor")
    selected_file = await save_file_dialog(
        "title",
        Path("testfile.gaphor"),
        filters=[],
    )

    assert file_dialog.calls["save"] == 1
    assert selected_file.parts == (os.path.sep, "test", "path", "model.gaphor")


@pytest.mark.asyncio
async def test_save_dialog_with_full_file_name(file_dialog):
    filename = Path("/test/path/model.gaphor")
    await save_file_dialog(
        "title",
        filename,
        filters=[],
    )

    assert file_dialog.get_initial_name() == filename.name


@pytest.mark.asyncio
async def test_save_dialog_filename_without_extension(file_dialog):
    file_dialog.define_response("/test/path/model")
    selected_file = await save_file_dialog(
        "title",
        Path("model"),
        filters=[],
    )

    assert selected_file.parts == (os.path.sep, "test", "path", "model")


@pytest.mark.asyncio
async def test_open_dialog_single_cancelled(file_dialog):
    file_dialog.define_response("/test/path/file")
    file_dialog.define_error(True)
    selected_file = await open_file_dialog("title", multiple=False)

    assert file_dialog.calls["open"] == 1
    assert selected_file is None


@pytest.mark.asyncio
async def test_open_dialog_one_file_single(file_dialog):
    file_dialog.define_response("/test/path/file")
    selected_file = await open_file_dialog(
        "title",
        multiple=False,
    )

    assert file_dialog.calls["open"] == 1
    assert isinstance(selected_file, Path)
    assert selected_file.parts == (os.path.sep, "test", "path", "file")


@pytest.mark.asyncio
async def test_open_dialog_multiple_cancelled(file_dialog):
    file_dialog.define_response(["/test/path/file"])
    file_dialog.define_error(True)
    selected_file = await open_file_dialog(
        "title",
    )

    assert file_dialog.calls["open_multiple"] == 1
    assert selected_file is None


@pytest.mark.asyncio
async def test_open_dialog_one_file_multiple(file_dialog):
    file_dialog.define_response(["/test/path/file"])
    selected_file = await open_file_dialog(
        "title",
    )

    assert file_dialog.calls["open_multiple"] == 1
    assert isinstance(selected_file, list)
    assert len(selected_file) == 1
    assert isinstance(selected_file[0], Path)
    assert selected_file[0].parts == (os.path.sep, "test", "path", "file")


def test_format_home_folder():
    display_name = pretty_path(Path.home() / "folder" / "mymodel.gaphor")

    assert display_name.startswith("~")
    assert display_name.endswith("mymodel.gaphor")


@pytest.mark.skipif(not hasattr(os, "getuid"), reason="Only on Unix")
def test_format_portal_document():
    display_name = pretty_path(
        Path(f"/run/user/{os.getuid()}/doc/123abc/mymodel.gaphor")
    )

    assert display_name.startswith("mymodel.gaphor")
    assert display_name.endswith(")")



import pytest

from gaphor.tests.raises import raises_exception_group


def test_capture_expected_exception():
    with raises_exception_group(ValueError):
        raise ExceptionGroup("msg", [ValueError()])


def test_capture_expected_from_multiple_exceptions():
    with raises_exception_group(ValueError):
        raise ExceptionGroup("msg", [TypeError(), ValueError()])


def test_propagate_unexpected_exception():
    with pytest.raises(ExceptionGroup):
        with raises_exception_group(ValueError):
            raise ExceptionGroup("msg", [TypeError()])


def test_propagate_wrong_exception():
    with pytest.raises(ValueError):
        with raises_exception_group(ValueError):
            raise ValueError()



import pytest
from gi.repository import Gdk

from gaphor import UML
from gaphor.core.modeling import StyleSheet
from gaphor.diagram.general import Line
from gaphor.diagram.styleeditor import (
    StyleEditor,
    StylePropertyPage,
    from_gdk_rgba,
    to_gdk_rgba,
)
from gaphor.diagram.tests.fixtures import find
from gaphor.UML.classes import ClassItem


@pytest.fixture
def style_sheet(element_factory):
    assert not next(element_factory.select(StyleSheet), None)
    return element_factory.create(StyleSheet)


@pytest.fixture
def style_editor(create, style_sheet, event_manager):
    item = create(ClassItem, UML.Class)
    style_editor = StyleEditor(item, style_sheet, event_manager, None, lambda: None)
    style_editor.present()
    yield style_editor
    style_editor.close()


class MockMainWindow:
    @property
    def window(self):
        return None


def test_style_editor_page(diagram, event_manager, element_factory):
    item = diagram.create(Line)
    property_page = StylePropertyPage(
        item, event_manager, element_factory, MockMainWindow()
    )
    widget = property_page.construct()
    assert widget
    style_editor = find(widget, "style-editor")
    assert style_editor


def test_style_editor_creation(style_editor):
    color_picker = find(style_editor.window, "color")
    border_radius = find(style_editor.window, "border-radius")
    background_color = find(style_editor.window, "background-color")
    text_color = find(style_editor.window, "text-color")

    assert color_picker
    assert border_radius
    assert background_color
    assert text_color


def test_color_from_style_sheet(style_editor, style_sheet: StyleSheet):
    color_picker = find(style_editor.window, "color")
    border_radius = find(style_editor.window, "border-radius")
    background_color = find(style_editor.window, "background-color")
    text_color = find(style_editor.window, "text-color")

    assert border_radius.get_value() == 0.0
    assert from_gdk_rgba(color_picker.get_rgba()) == "rgba(0, 0, 0, 1.0)"
    assert from_gdk_rgba(background_color.get_rgba()) == "rgba(0, 0, 0, 0.0)"
    assert from_gdk_rgba(text_color.get_rgba()) == "rgba(0, 0, 0, 1.0)"


def test_set_border_radius(style_editor, style_sheet):
    border_radius = find(style_editor.window, "border-radius")
    border_radius.set_value(10)

    assert "border-radius: 10.0" in style_sheet.instant_style_declarations


def test_set_color(style_editor, style_sheet: StyleSheet):
    color_picker = find(style_editor.window, "color")
    color_picker.set_rgba(to_gdk_rgba("rgba(255, 0, 0, 1)"))
    style_editor.on_color_set(color_picker)

    assert "color: rgba(255, 0, 0, 1.0)" in style_sheet.instant_style_declarations


def test_set_background_color(style_editor, style_sheet):
    background_color = find(style_editor.window, "background-color")
    rgba = Gdk.RGBA()
    rgba.parse("rgba(0, 255, 0, 1)")
    background_color.set_rgba(rgba)
    style_editor.on_background_color_set(background_color)

    assert (
        "background-color: rgba(0, 255, 0, 1.0)"
        in style_sheet.instant_style_declarations
    )


def test_set_text_color(style_editor, style_sheet):
    text_color = find(style_editor.window, "text-color")
    rgba = Gdk.RGBA()
    rgba.parse("rgba(0, 0, 255, 1)")
    text_color.set_rgba(rgba)
    style_editor.on_text_color_set(text_color)

    assert "text-color: rgba(0, 0, 255, 1.0)" in style_sheet.instant_style_declarations


def test_change_name(style_editor, style_sheet):
    item = style_editor.subject
    item.subject.name = "Name"
    color = find(style_editor.window, "color")
    rgba = Gdk.RGBA()
    rgba.parse("rgba(0, 0, 255, 1)")
    color.set_rgba(rgba)
    style_editor.on_color_set(color)
    assert (
        'class[name="Name"] {\n  color: rgba(0, 0, 255, 1.0)'
        in style_sheet.instant_style_declarations
    )

    item.subject.name = "New Name"

    assert (
        'class[name="New Name"] {\n  color: rgba(0, 0, 255, 1.0)'
        in style_sheet.instant_style_declarations
    )


def test_element_deleted(style_editor, style_sheet):
    item = style_editor.subject
    item.subject.name = "Name"
    color = find(style_editor.window, "color")
    rgba = Gdk.RGBA()
    rgba.parse("rgba(0, 0, 255, 1)")
    color.set_rgba(rgba)
    style_editor.on_color_set(color)
    assert (
        'class[name="Name"] {\n  color: rgba(0, 0, 255, 1.0)'
        in style_sheet.instant_style_declarations
    )

    item.unlink()

    assert style_sheet.instant_style_declarations == ""


def test_export(style_sheet, style_editor, diagram, element_factory, event_manager):
    item = style_editor.subject
    item.subject.name = "Name"

    color = find(style_editor.window, "color")
    rgba = Gdk.RGBA()
    rgba.parse("rgba(0, 0, 255, 1)")
    color.set_rgba(rgba)
    style_editor.on_color_set(color)

    export = find(style_editor.window, "export")
    style_editor.on_export(export)

    assert "rgba(0, 0, 255, 1.0)" in style_sheet.styleSheet



from gaphor.settings import get_cache_dir, get_config_dir


def test_config_dir():
    config_dir = get_config_dir()
    assert str(config_dir).endswith("gaphor")


def test_cache_dir():
    cache_dir = get_cache_dir()
    assert str(cache_dir).endswith("gaphor")



# A workaround for Pydot, to make sure it works
# in a PyInstaller setup.
# The application path is updated and the
# environment variable VGBINDIR is set to make sure plugins are found

# ruff: noqa: SLF001

import logging
import os
import subprocess
import sys

import pydot.core

log = logging.getLogger(__name__)


def graphviz_path():
    if hasattr(sys, "_MEIPASS"):
        # Use packaged `dot` (pyinstaller)
        if sys.platform == "win32":
            return os.path.join(sys._MEIPASS, "graphviz")
        else:
            return sys._MEIPASS
    return None


def graphviz_plugin_path():
    if hasattr(sys, "_MEIPASS"):
        return os.path.join(sys._MEIPASS, "graphviz")
    return None


def have_graphviz(program):
    if path := graphviz_path():
        program = os.path.join(path, program + pydot.get_executable_extension())
    return os.system(f'"{program}" -V') == 0


def call_graphviz(program, arguments, working_dir, **kwargs):
    if path := graphviz_path():
        program = os.path.join(path, program + pydot.get_executable_extension())

    if arguments is None:
        arguments = []

    plugin_path = graphviz_plugin_path()
    env = {
        "PATH": os.environ.get("PATH", ""),
        "LD_LIBRARY_PATH": os.environ.get("LD_LIBRARY_PATH", ""),
        "SYSTEMROOT": os.environ.get("SYSTEMROOT", ""),
        **({"VGBINDIR": plugin_path} if plugin_path else {}),
    }

    program_with_args = [program] + arguments

    process = subprocess.Popen(
        program_with_args,
        env=env,
        cwd=working_dir,
        shell=False,
        stderr=subprocess.PIPE,
        stdout=subprocess.PIPE,
        **kwargs,
    )
    stdout_data, stderr_data = process.communicate()

    return stdout_data, stderr_data, process


pydot.core.call_graphviz = call_graphviz



from gi.repository import GLib

from gaphor.core.modeling import ModelReady
from gaphor.ui.modelchanged import ModelChanged
from gaphor.ui.tests.fixtures import iterate_until


def iteration(sentinel):
    ctx = GLib.main_context_default()
    while ctx.pending() and not sentinel:
        ctx.iteration(False)


def test_create_ui(event_manager):
    model_changed = ModelChanged(event_manager)

    widget = model_changed.open()

    assert widget


def test_monitor_file_changed(event_manager, tmp_path):
    new_file = tmp_path / "new_file"
    new_file.write_text("a", encoding="utf-8")

    model_changed = ModelChanged(event_manager)
    widget = model_changed.open()
    event_manager.handle(ModelReady(None, filename=new_file))

    new_file.write_text("b", encoding="utf-8")
    iterate_until(widget.get_revealed)

    assert widget.get_revealed()


def test_monitor_file_not_changed(event_manager, tmp_path):
    new_file = tmp_path / "new_file"
    new_file.write_text("a", encoding="utf-8")

    model_changed = ModelChanged(event_manager)
    widget = model_changed.open()
    event_manager.handle(ModelReady(None, filename=new_file))

    new_file.write_text("a", encoding="utf-8")
    iterate_until(condition=lambda: False, timeout=2)

    assert not widget.get_revealed()



import pytest

from gaphor.core.modeling import StyleSheet
from gaphor.storage.load import load_elements
from gaphor.storage.parser import element


@pytest.fixture
def loader(element_factory, modeling_language):
    def _loader(*parsed_elements):
        parsed_data = {p.id: p for p in parsed_elements}
        load_elements(parsed_data, element_factory, modeling_language)
        *elements, style_sheet = element_factory.lselect()
        assert isinstance(style_sheet, StyleSheet)
        return elements

    return _loader


def test_owned_comment_to_comment_upgrade(loader):
    c = element(id="1", type="Comment")
    c.references["annotatedElement"] = ["2"]
    e = element(id="2", type="Element")
    e.references["ownedComment"] = ["1"]

    comment, elem = loader(c, e)
    assert elem in comment.annotatedElement
    assert comment in elem.comment


def test_owned_classifier_to_owned_type(loader):
    p = element(id="1", type="Package")
    p.references["ownedClassifier"] = ["2"]
    c = element(id="2", type="Class")

    package, clazz = loader(p, c)
    assert clazz in package.ownedType
    assert clazz.package is package


def test_implementation_to_interface_realization(loader):
    i = element(id="1", type="InterfaceRealization")
    c = element(id="2", type="Class")
    c.references["interfaceRealization"] = ["1"]

    interface_realization, clazz = loader(i, c)
    assert interface_realization in clazz.clientDependency
    assert clazz is interface_realization.implementingClassifier


def test_formal_parameter_to_owned_parameter(loader):
    o = element(id="1", type="Operation")
    o.references["formalParameter"] = ["2"]
    p = element(id="2", type="Parameter")

    operation, parameter = loader(o, p)
    assert parameter in operation.ownedParameter
    assert parameter.operation is operation


def test_return_result_to_owned_parameter(loader):
    o = element(id="1", type="Operation")
    o.references["returnResult"] = ["2"]
    p = element(id="2", type="Parameter")
    p.references["ownerReturnParam"] = ["1"]

    operation, parameter = loader(o, p)
    assert parameter in operation.ownedParameter
    assert parameter.operation is operation


def test_parameters_to_owned_parameter(loader):
    o = element(id="1", type="Operation")
    o.references["formalParameter"] = ["2"]
    o.references["returnResult"] = ["3"]
    p = element(id="2", type="Parameter")
    r = element(id="3", type="Parameter")

    operation, parameter, return_param = loader(o, p, r)
    assert parameter in operation.ownedParameter
    assert return_param in operation.ownedParameter
    assert parameter.operation is operation
    assert return_param.operation is operation



"""A Property-based test.

This is a property based/model based/monkey test.

It starts a user session and performs all sorts of user
actions:
- create/delete elements
- create/delete diagrams
- connect, disconnect
- change owner element
- undo, redo
- copy, paste

Some tips:
- the model is leading. Just draw from the model with the proper filters
- do not perform `assume()` calls in a transaction

To do:
- Create tests for actions, interactions, states
- Create tests for SysML tools (combined with UML?)
"""

from __future__ import annotations

import contextlib
import itertools
from collections.abc import Iterable
from functools import singledispatch
from io import StringIO

from gaphas.connector import Handle
from hypothesis import reproduce_failure, settings  # noqa
from hypothesis.control import assume, cleanup
from hypothesis.errors import UnsatisfiedAssumption
from hypothesis.stateful import (
    RuleBasedStateMachine,
    initialize,
    invariant,
    rule,
)
from hypothesis.strategies import data, integers, lists, sampled_from

import gaphor.storage as storage
from gaphor import UML
from gaphor.application import Session
from gaphor.C4Model.toolbox import c4
from gaphor.core import Transaction
from gaphor.core.modeling import (
    Base,
    Diagram,
    ElementFactory,
    Presentation,
    StyleSheet,
)
from gaphor.core.modeling.base import generate_id, uuid_generator
from gaphor.diagram.copypaste import copy_full, paste_full, paste_link
from gaphor.diagram.deletable import deletable
from gaphor.diagram.drop import drop
from gaphor.diagram.group import can_group, change_owner
from gaphor.diagram.presentation import LinePresentation
from gaphor.diagram.support import get_diagram_item_metadata
from gaphor.diagram.tests.fixtures import allow, connect, disconnect
from gaphor.RAAML.toolbox import fta, stpa
from gaphor.SysML.toolbox import blocks, internal_blocks, requirements
from gaphor.UML import diagramitems
from gaphor.UML.toolbox import (
    actions,
    classes,
    deployments,
    interactions,
    profiles,
    states,
    use_cases,
)


def load_default_model(element_factory):
    element_factory.flush()
    with element_factory.block_events():
        element_factory.create(StyleSheet)
        model = element_factory.create(UML.Package)
        model.name = "New model"
        diagram = element_factory.create(UML.Diagram)
        diagram.element = model
        diagram.name = "New diagram"


def tooldef():
    return sampled_from(
        list(
            itertools.chain(
                # UML
                classes.tools,
                deployments.tools,
                use_cases.tools,
                profiles.tools,
                actions.tools,
                interactions.tools,
                states.tools,
                # C4
                c4.tools,
                # SysML
                blocks.tools,
                internal_blocks.tools,
                requirements.tools,
                # RAAML
                stpa.tools,
                fta.tools,
            )
        )
    )


class ModelConsistency(RuleBasedStateMachine):
    @property
    def model(self) -> ElementFactory:
        return self.session.get_service("element_factory")  # type: ignore[no-any-return]

    @property
    def transaction(self) -> Transaction:
        return Transaction(self.session.get_service("event_manager"))

    def select(self, expression=None):
        elements = ordered(self.model.select(expression))
        assume(elements)
        return sampled_from(elements)

    def diagrams(self):
        return self.select(lambda e: isinstance(e, UML.Diagram))

    def relations(self, diagram):
        relations = [
            p
            for p in diagram.presentation
            if isinstance(p, diagramitems.DependencyItem)
        ]
        assume(relations)
        return sampled_from(ordered(relations))

    def targets(self, relation, handle):
        return self.select(
            lambda e: (
                isinstance(e, diagramitems.ClassItem)
                and e.diagram is relation.diagram
                and allow(relation, handle, e)
            )
        )

    @initialize()
    def new_session(self):
        generate_id(map(str, itertools.count()))
        cleanup(lambda: generate_id(uuid_generator()))

        self.session = Session()
        self.copy_buffer: object = None

        load_default_model(self.model)
        self.fully_pasted_items = set()

    @rule()
    def create_diagram(self):
        with self.transaction:
            self.model.create(Diagram)

    @rule(
        tooldef=tooldef(),
        data=data(),
        x=integers(min_value=0, max_value=2000),
        y=integers(min_value=0, max_value=2000),
    )
    def add_item_to_diagram(self, tooldef, data, x, y):
        diagram = data.draw(self.diagrams())
        with self.transaction:
            item = tooldef.item_factory(diagram)
            item.matrix.translate(x, y)
            diagram.update({item})

        # Do best effort to connect a line, no problem if it fails
        if isinstance(item, LinePresentation):
            self.try_connect_relation(data, item, item.head)
            self.try_connect_relation(data, item, item.tail)

    def try_connect_relation(self, data, item, handle):
        with contextlib.suppress(UnsatisfiedAssumption):
            self._connect_relation(data, item, handle)

    @rule(data=data())
    def delete_element(self, data):
        # Do not delete StyleSheet: it will be re-created on load,
        # causing test invariants to fail. It can't be created
        # dynamically, because such changes require a transaction.
        elements = self.select(lambda e: not isinstance(e, StyleSheet) and deletable(e))
        element = data.draw(elements)
        with self.transaction:
            element.unlink()

    @rule(data=data())
    def connect_relation(self, data):
        diagram = data.draw(self.diagrams())
        relation = data.draw(self.relations(diagram))
        handle = data.draw(sampled_from([relation.head, relation.tail]))
        self._connect_relation(data, relation, handle)

    def _connect_relation(self, data, relation, handle):
        target = data.draw(self.targets(relation, handle))
        with self.transaction:
            connect(relation, handle, target)

    @rule(data=data())
    def disconnect_relation(self, data):
        diagram = data.draw(self.diagrams())
        relation = data.draw(self.relations(diagram))
        handle = data.draw(sampled_from([relation.head, relation.tail]))
        with self.transaction:
            disconnect(relation, handle)

    @rule(data=data())
    def change_owner(self, data):
        parent = data.draw(self.select())
        element = data.draw(
            self.select(lambda e: e is not parent and can_group(parent, e))
        )
        with self.transaction as tx:
            changed = change_owner(parent, element)
            if not changed:
                tx.rollback()
        assume(changed)

    @rule(data=data())
    def drop(self, data):
        diagram = data.draw(self.diagrams())
        element = data.draw(self.select())
        with self.transaction as tx:
            item = drop(element, diagram, 0, 0)
            if not item:
                tx.rollback()
        assume(item)

    @rule()
    def undo(self):
        undo_manager = self.session.get_service("undo_manager")
        assume(undo_manager.can_undo())
        undo_manager.undo_transaction()

    @rule()
    def redo(self):
        undo_manager = self.session.get_service("undo_manager")
        assume(undo_manager.can_redo())
        undo_manager.redo_transaction()

    @rule(data=data())
    def copy(self, data):
        diagram = data.draw(self.diagrams())
        assume(diagram.ownedPresentation)
        # Take from model, to ensure order.
        items = data.draw(
            lists(
                sampled_from(ordered(diagram.ownedPresentation)),
                min_size=1,
                unique=True,
            )
        )
        self.copy_buffer = copy_full(items)

    @rule(data=data())
    def paste_link(self, data):
        assume(self.copy_buffer)
        diagram = data.draw(self.diagrams())
        with self.transaction:
            paste_link(self.copy_buffer, diagram)

    @rule(data=data())
    def paste_full(self, data):
        assume(self.copy_buffer)
        diagram = data.draw(self.diagrams())
        with self.transaction:
            new_items = paste_full(self.copy_buffer, diagram)
        self.fully_pasted_items.update(new_items)

    @invariant()
    def check_relations(self):
        for relation in self.model.select(LinePresentation):
            subject = relation.subject
            diagram = relation.diagram
            head = get_connected(diagram, relation.head)
            tail = get_connected(diagram, relation.tail)

            if head and tail:
                check_relation(relation, head, tail)
            elif relation not in self.fully_pasted_items:
                assert not subject

    @invariant()
    def check_save_and_load(self):
        new_model = ElementFactory()
        try:
            with StringIO() as buffer:
                storage.save(buffer, self.model)
                buffer.seek(0)
                storage.load(
                    buffer,
                    element_factory=new_model,
                    modeling_language=self.session.get_service("modeling_language"),
                )

            assert self.model.size() == new_model.size(), (
                f"{self.model.lselect()} != {new_model.lselect()}"
            )
            assert {e.id for e in self.model} == {e.id for e in new_model}
        except Exception:
            with open("falsifying_model.gaphor", "w", encoding="utf") as out:
                storage.save(out, self.model)
            raise


def get_connected(diagram: Diagram, handle: Handle) -> Presentation | None:
    """Get item connected to a handle."""
    if cinfo := diagram.connections.get_connection(handle):
        return cinfo.connected  # type: ignore[no-any-return]
    return None


def ordered(elements: Iterable[Base]) -> list[Base]:
    return sorted(elements, key=lambda e: e.id)


@singledispatch
def check_relation(relation: Presentation, head: Presentation, tail: Presentation):
    subject = relation.subject
    assert subject

    metadata = get_diagram_item_metadata(type(relation))

    def end_elements(end):
        return (
            [metadata[end].get(subject)]
            if metadata[end].upper == 1
            else metadata[end].get(subject)
        )

    assert metadata, f"No comparison function for {relation}"
    assert head.subject in end_elements("head")
    assert tail.subject in end_elements("tail")


@check_relation.register
def _(relation: diagramitems.ContainmentItem, head, tail):
    assert not relation.subject
    # subject ownership can not be checked, since
    # it can be changed by the group functionality.


@check_relation.register
def _(relation: diagramitems.AssociationItem, head, tail):
    subject = relation.subject
    targets = [m.type for m in subject.memberEnd]
    assert head.subject in targets
    assert tail.subject in targets


@check_relation.register
def _(relation: diagramitems.ExtensionItem, head, tail):
    subject = relation.subject
    targets = [m.type for m in subject.memberEnd]
    assert head.subject in targets
    assert tail.subject in targets


@check_relation.register
def _(relation: diagramitems.MessageItem, head, tail):
    subject = relation.subject
    assert subject
    assert isinstance(subject.sendEvent, UML.MessageOccurrenceSpecification)
    assert isinstance(subject.receiveEvent, UML.MessageOccurrenceSpecification)
    assert subject.sendEvent.covered is head.subject
    assert subject.receiveEvent.covered is tail.subject


TestModelConsistency = ModelConsistency.TestCase



"""Flow item connection adapters tests."""

import pytest as pytest

from gaphor import UML
from gaphor.diagram.tests.fixtures import allow, connect, disconnect, get_connected
from gaphor.UML.actions.action import ActionItem
from gaphor.UML.actions.activity import ActivityItem, ActivityParameterNodeItem
from gaphor.UML.actions.activitynodes import (
    ActivityFinalNodeItem,
    DecisionNodeItem,
    FlowFinalNodeItem,
    ForkNodeItem,
    InitialNodeItem,
)
from gaphor.UML.actions.flow import ControlFlowItem, ObjectFlowItem
from gaphor.UML.actions.objectnode import ObjectNodeItem
from gaphor.UML.actions.pin import InputPinItem, OutputPinItem


def test_initial_node_glue(create):
    """Test flow item gluing to initial node item."""

    flow = create(ControlFlowItem)
    node = create(InitialNodeItem, UML.InitialNode)

    # tail may not connect to initial node item
    allowed = allow(flow, flow.tail, node)
    assert not allowed

    allowed = allow(flow, flow.head, node)
    assert allowed


def test_flow_final_node_glue(create):
    """Test flow item gluing to flow final node item."""

    flow = create(ControlFlowItem)
    node = create(FlowFinalNodeItem, UML.FlowFinalNode)

    # head may not connect to flow final node item
    allowed = allow(flow, flow.head, node)
    assert not allowed

    allowed = allow(flow, flow.tail, node)
    assert allowed


def test_activity_final_node_glue(create):
    """Test flow item gluing to activity final node item."""
    flow = create(ControlFlowItem)
    node = create(ActivityFinalNodeItem, UML.ActivityFinalNode)

    # head may not connect to activity final node item
    glued = allow(flow, flow.head, node)
    assert not glued

    glued = allow(flow, flow.tail, node)
    assert glued


def test_glue_to_object_node(create):
    flow = create(ObjectFlowItem)
    onode = create(ObjectNodeItem, UML.ObjectNode)
    glued = allow(flow, flow.head, onode)
    assert glued


def test_glue_to_activity_parameter_node(create, element_factory):
    flow = create(ObjectFlowItem)
    anode = create(ActivityParameterNodeItem, UML.ActivityParameterNode)

    glued = allow(flow, flow.head, anode)

    assert glued


def test_connect_to_object_node(create):
    flow = create(ObjectFlowItem)
    anode = create(ActionItem, UML.Action)
    onode = create(ObjectNodeItem, UML.ObjectNode)

    connect(flow, flow.head, anode)
    connect(flow, flow.tail, onode)
    assert flow.subject
    assert isinstance(flow.subject, UML.ObjectFlow)

    disconnect(flow, flow.head)
    disconnect(flow, flow.tail)

    # opposite connection
    connect(flow, flow.head, onode)
    connect(flow, flow.tail, anode)
    assert flow.subject
    assert isinstance(flow.subject, UML.ObjectFlow)


def test_connect_to_activity_parameter_node(create, element_factory):
    flow = create(ObjectFlowItem)
    activity = create(ActivityItem, UML.Activity)
    param_node = element_factory.create(UML.ActivityParameterNode)
    param_node.parameter = element_factory.create(UML.Parameter)
    activity.subject.node = param_node
    anode = param_node.presentation[0]
    onode = create(ObjectNodeItem, UML.ObjectNode)

    connect(flow, flow.head, anode)
    connect(flow, flow.tail, onode)
    assert flow.subject
    assert isinstance(flow.subject, UML.ObjectFlow)


def test_do_not_allow_head_to_input_pin(create):
    flow = create(ObjectFlowItem)
    pin = create(InputPinItem, UML.InputPin)

    assert not allow(flow, flow.head, pin)
    assert allow(flow, flow.tail, pin)


def test_connect_tail_to_input_pin(create):
    flow = create(ObjectFlowItem)
    pin = create(InputPinItem, UML.InputPin)
    onode = create(ObjectNodeItem, UML.ObjectNode)

    connect(flow, flow.head, onode)
    connect(flow, flow.tail, pin)

    assert flow.subject
    assert isinstance(flow.subject, UML.ObjectFlow)


def test_do_not_allow_tail_to_output_pin(create):
    flow = create(ObjectFlowItem)
    pin = create(OutputPinItem, UML.OutputPin)

    assert allow(flow, flow.head, pin)
    assert not allow(flow, flow.tail, pin)


def test_connect_head_to_output_pin(create):
    flow = create(ObjectFlowItem)
    pin = create(OutputPinItem, UML.OutputPin)
    onode = create(ObjectNodeItem, UML.ObjectNode)

    connect(flow, flow.head, pin)
    connect(flow, flow.tail, onode)

    assert flow.subject
    assert isinstance(flow.subject, UML.ObjectFlow)


def test_object_flow_activity_is_set_on_output_pin(create):
    activity = create(ActivityItem, UML.Activity)
    action = create(ActionItem, UML.Action)
    action.subject.activity = activity.subject

    out_pin = create(OutputPinItem, UML.OutputPin)
    connect(out_pin, out_pin.handles()[0], action)

    in_pin = create(InputPinItem, UML.InputPin)
    flow = create(ObjectFlowItem)
    connect(flow, flow.head, out_pin)
    connect(flow, flow.tail, in_pin)

    assert flow.subject.owner is activity.subject


def test_object_flow_activity_is_set_on_input_pin(create):
    activity = create(ActivityItem, UML.Activity)
    action = create(ActionItem, UML.Action)
    action.subject.activity = activity.subject

    in_pin = create(InputPinItem, UML.InputPin)
    connect(in_pin, in_pin.handles()[0], action)

    out_pin = create(OutputPinItem, UML.OutputPin)
    flow = create(ObjectFlowItem)
    connect(flow, flow.head, out_pin)
    connect(flow, flow.tail, in_pin)

    assert flow.subject.owner is activity.subject


def test_object_flow_reconnect(create, element_factory, sanitizer_service):
    flow = create(ObjectFlowItem)
    a1 = create(ActionItem, UML.Action)
    o1 = create(ObjectNodeItem, UML.ObjectNode)
    o2 = create(ObjectNodeItem, UML.ObjectNode)

    # connect: a1 -> o1
    connect(flow, flow.head, a1)
    connect(flow, flow.tail, o1)

    f = flow.subject
    f.name = "tname"
    guard = f.model.create(UML.LiteralString)
    guard.value = "tguard"
    guard.activityEdge = f
    f.guard = guard

    # reconnect: a1 -> o2
    connect(flow, flow.tail, o2)

    assert len(a1.subject.incoming) == 0
    assert len(a1.subject.outgoing) == 1
    # no connections to o1
    assert len(o1.subject.incoming) == 0
    assert len(o1.subject.outgoing) == 0
    # connections to o2 instead
    assert len(o2.subject.incoming) == 1
    assert len(o2.subject.outgoing) == 0

    assert len(element_factory.lselect(UML.ObjectFlow)) == 1
    # one guard
    assert flow.subject.name == "tname"
    assert flow.subject.guard.value == "tguard"


def test_control_flow_reconnection(create, sanitizer_service):
    """Test control flow becoming object flow due to reconnection."""
    flow = create(ControlFlowItem)
    a1 = create(ActionItem, UML.Action)
    a2 = create(ActionItem, UML.Action)
    a3 = create(ActionItem, UML.Action)

    # connect with control flow: a1 -> a2
    connect(flow, flow.head, a1)
    connect(flow, flow.tail, a2)

    f = flow.subject
    f.name = "tname"
    guard = f.model.create(UML.LiteralString)
    guard.value = "tguard"
    guard.activityEdge = f
    f.guard = guard

    # reconnect with object flow: a1 -> o1
    connect(flow, flow.tail, a3)

    assert len(a1.subject.incoming) == 0
    assert len(a1.subject.outgoing) == 1
    # no connections to a2
    assert len(a2.subject.incoming) == 0
    assert len(a2.subject.outgoing) == 0
    # connections to o1 instead
    assert len(a3.subject.incoming) == 1
    assert len(a3.subject.outgoing) == 0

    # one guard, not changed
    assert flow.subject.name == "tname"
    assert flow.subject.guard.value == "tguard"


def test_glue(create):
    """Test flow item gluing to action items."""

    flow = create(ControlFlowItem)
    a1 = create(ActionItem, UML.Action)
    a2 = create(ActionItem, UML.Action)

    glued = allow(flow, flow.head, a1)
    assert glued

    connect(flow, flow.head, a1)

    glued = allow(flow, flow.tail, a2)
    assert glued


def test_connect_to_action_item(create):
    flow = create(ControlFlowItem)
    a1 = create(ActionItem, UML.Action)
    a2 = create(ActionItem, UML.Action)

    connect(flow, flow.head, a1)
    connect(flow, flow.tail, a2)

    assert isinstance(flow.subject, UML.ControlFlow)

    assert len(a1.subject.incoming) == 0
    assert len(a2.subject.incoming) == 1
    assert len(a1.subject.outgoing) == 1
    assert len(a2.subject.outgoing) == 0

    assert flow.subject in a1.subject.outgoing
    assert flow.subject.source is a1.subject
    assert flow.subject in a2.subject.incoming
    assert flow.subject.target is a2.subject


def test_disconnect_from_action_item(create, sanitizer_service):
    """Test flow item disconnection from action items."""
    flow = create(ControlFlowItem)
    a1 = create(ActionItem, UML.Action)
    a2 = create(ActionItem, UML.Action)

    connect(flow, flow.head, a1)
    connect(flow, flow.tail, a2)

    disconnect(flow, flow.head)
    assert flow.subject is None
    assert len(a1.subject.incoming) == 0
    assert len(a2.subject.incoming) == 0
    assert len(a1.subject.outgoing) == 0
    assert len(a2.subject.outgoing) == 0


def test_reconnect(create, element_factory, sanitizer_service):
    """Test flow item reconnection."""
    flow = create(ControlFlowItem)
    a1 = create(ActionItem, UML.Action)
    a2 = create(ActionItem, UML.Action)
    a3 = create(ActionItem, UML.Action)

    # a1 -> a2
    connect(flow, flow.head, a1)
    connect(flow, flow.tail, a2)
    f = flow.subject
    f.name = "tname"
    guard = f.model.create(UML.LiteralString)
    guard.value = "tguard"
    guard.activityEdge = f
    f.guard = guard

    # reconnect: a1 -> a3
    connect(flow, flow.tail, a3)

    assert len(a1.subject.incoming) == 0
    assert len(a1.subject.outgoing) == 1
    # no connections to a2
    assert len(a2.subject.incoming) == 0
    assert len(a2.subject.outgoing) == 0
    # connections to a3 instead
    assert len(a3.subject.incoming) == 1
    assert len(a3.subject.outgoing) == 0

    assert len(element_factory.lselect(UML.ControlFlow)) == 1
    # one guard
    assert flow.subject.name == "tname"
    assert flow.subject.guard.value == "tguard"


def test_object_flow_reconnection(create, sanitizer_service):
    """Test object flow becoming control flow due to reconnection."""
    flow = create(ObjectFlowItem)
    a1 = create(ActionItem, UML.Action)
    o1 = create(ObjectNodeItem, UML.ObjectNode)
    o2 = create(ObjectNodeItem, UML.ObjectNode)

    # connect with control flow: a1 -> o1
    connect(flow, flow.head, a1)
    connect(flow, flow.tail, o1)

    f = flow.subject
    f.name = "tname"
    guard = f.model.create(UML.LiteralString)
    guard.value = "tguard"
    guard.activityEdge = f
    f.guard = guard

    # reconnect with object flow: a1 -> a2
    connect(flow, flow.tail, o2)

    assert len(a1.subject.incoming) == 0
    assert len(a1.subject.outgoing) == 1
    # no connections to o1
    assert len(o1.subject.incoming) == 0
    assert len(o1.subject.outgoing) == 0
    # connections to a2 instead
    assert len(o2.subject.incoming) == 1
    assert len(o2.subject.outgoing) == 0

    # one guard, not changed
    assert flow.subject.name == "tname"
    assert flow.subject.guard.value == "tguard"


fork_and_decision_items = pytest.mark.parametrize(
    "item_cls,fork_node_cls,join_node_cls",
    [
        [ForkNodeItem, UML.ForkNode, UML.JoinNode],
        [DecisionNodeItem, UML.DecisionNode, UML.MergeNode],
    ],
)


@fork_and_decision_items
@pytest.mark.parametrize("flow_item", [ControlFlowItem, ObjectFlowItem])
def test_glue_flow_item(create, item_cls, fork_node_cls, join_node_cls, flow_item):
    """Test decision/fork nodes glue."""
    flow = create(flow_item)
    action = create(ActionItem, UML.Action)
    node = create(item_cls, join_node_cls)

    glued = allow(flow, flow.head, node)
    assert glued

    connect(flow, flow.head, action)

    glued = allow(flow, flow.tail, node)
    assert glued


@fork_and_decision_items
@pytest.mark.parametrize("flow_item", [ControlFlowItem, ObjectFlowItem])
def test_node_class_change(create, item_cls, fork_node_cls, join_node_cls, flow_item):
    """Test node incoming edges.

    Connection scheme is presented below::

                head  tail
        [ a1 ]--flow1-->    |
                            [ jn ] --flow3--> [ a3 ]
        [ a2 ]--flow2-->    |

    Node class changes due to two incoming edges and one outgoing edge.
    """
    flow1 = create(flow_item)
    flow2 = create(flow_item)
    flow3 = create(flow_item)
    a1 = create(ActionItem, UML.Action)
    a2 = create(ActionItem, UML.Action)
    jn = create(item_cls, fork_node_cls)

    assert isinstance(jn.subject, fork_node_cls)

    # connect actions first
    connect(flow1, flow1.head, a1)
    connect(flow2, flow2.head, a2)
    connect(flow3, flow3.tail, a2)

    # connect to the node
    connect(flow1, flow1.tail, jn)
    connect(flow2, flow2.tail, jn)
    connect(flow3, flow3.head, jn)

    # node class changes
    assert isinstance(jn.subject, join_node_cls)


@fork_and_decision_items
@pytest.mark.parametrize("flow_item", [ControlFlowItem, ObjectFlowItem])
def test_outgoing_edges(create, item_cls, fork_node_cls, join_node_cls, flow_item):
    """Test outgoing edges.

    Connection scheme is presented below::

                head  tail    | --flow2-->[ a2 ]
        [ a1 ] --flow1--> [ jn ]
                                | --flow3-->[ a3 ]
    """
    flow1 = create(flow_item)
    flow2 = create(flow_item)
    flow3 = create(flow_item)
    a1 = create(ActionItem, UML.Action)
    a2 = create(ActionItem, UML.Action)
    jn = create(item_cls, join_node_cls)

    # connect actions first
    connect(flow1, flow1.head, a1)
    connect(flow2, flow2.tail, a2)
    connect(flow3, flow3.tail, a2)

    # connect to the node
    connect(flow1, flow1.tail, jn)
    assert isinstance(jn.subject, join_node_cls)

    connect(flow2, flow2.head, jn)
    assert isinstance(jn.subject, join_node_cls)

    assert len(jn.subject.incoming) == 1
    assert len(jn.subject.outgoing) == 1
    assert flow1.subject in jn.subject.incoming
    assert flow2.subject in jn.subject.outgoing

    connect(flow3, flow3.head, jn)
    assert len(jn.subject.outgoing) == 2

    assert isinstance(jn.subject, fork_node_cls), f"{jn.subject}"


@fork_and_decision_items
@pytest.mark.parametrize("flow_item", [ControlFlowItem, ObjectFlowItem])
def test_combined_nodes_connection(
    create, item_cls, fork_node_cls, join_node_cls, flow_item
):
    """Test combined nodes connection.

    Connection scheme is presented below::

                head  tail    |   --flow2--> [ a2 ]
        [ a1 ] --flow1--> [ jn ]
        [ a4 ] --flow4-->    |   --flow3--> [ a3 ]

    Flow `flow4` will force the node to become a combined node.
    """
    flow1 = create(flow_item)
    flow2 = create(flow_item)
    flow3 = create(flow_item)
    flow4 = create(flow_item)
    a1 = create(ActionItem, UML.Action)
    a2 = create(ActionItem, UML.Action)
    a4 = create(ActionItem, UML.Action)
    jn = create(item_cls, join_node_cls)

    # connect actions first
    connect(flow1, flow1.head, a1)
    connect(flow2, flow2.tail, a2)
    connect(flow3, flow3.tail, a2)
    connect(flow4, flow4.head, a4)

    # connect to the node
    connect(flow1, flow1.tail, jn)
    connect(flow2, flow2.head, jn)
    connect(flow3, flow3.head, jn)

    connect(flow4, flow4.tail, jn)
    assert isinstance(jn.subject, join_node_cls)
    assert jn.combined is not None

    # check node combination
    assert len(jn.subject.outgoing) == 1
    assert len(jn.combined.incoming) == 1
    assert jn.subject.outgoing[0] is jn.combined.incoming[0]


@fork_and_decision_items
@pytest.mark.parametrize(
    "flow_item,uml_flow",
    [(ControlFlowItem, UML.ControlFlow), (ObjectFlowItem, UML.ObjectFlow)],
)
def test_combined_node_disconnection(
    create,
    element_factory,
    item_cls,
    fork_node_cls,
    join_node_cls,
    flow_item,
    uml_flow,
    sanitizer_service,
):
    """Test combined nodes disconnection.

    Connection scheme is presented below::

                head  tail    |   --flow2--> [ a2 ]
        [ a1 ] --flow1--> [ jn ]
        [ a4 ] --flow4-->    |   --flow3--> [ a3 ]

    Flow `flow4` will force the node to become a combined node.
    """
    flow1 = create(flow_item)
    flow2 = create(flow_item)
    flow3 = create(flow_item)
    flow4 = create(flow_item)
    a1 = create(ActionItem, UML.Action)
    a2 = create(ActionItem, UML.Action)
    a4 = create(ActionItem, UML.Action)
    jn = create(item_cls, join_node_cls)

    # connect actions first
    connect(flow1, flow1.head, a1)
    connect(flow2, flow2.tail, a2)
    connect(flow3, flow3.tail, a2)
    connect(flow4, flow4.head, a4)

    # connect to the node
    connect(flow1, flow1.tail, jn)
    connect(flow2, flow2.head, jn)
    connect(flow3, flow3.head, jn)
    connect(flow4, flow4.tail, jn)

    # needed for tests below
    flow = jn.subject.outgoing[0]
    node = jn.combined

    assert flow in element_factory.lselect(uml_flow)
    assert node in element_factory.lselect(fork_node_cls)

    # test disconnection
    disconnect(flow4, flow4.head)
    assert get_connected(flow4, flow4.head) is None

    assert jn.combined is None

    flows = element_factory.lselect(uml_flow)
    nodes = element_factory.lselect(fork_node_cls)
    assert node not in nodes, f"{node} in {nodes}"
    assert flow not in flows, f"{flow} in {flows}"



from gaphas.canvas import instant_cairo_context

from gaphor import UML
from gaphor.core.modeling import DrawContext
from gaphor.UML.actions.flow import ControlFlowItem


def test_flow(diagram, element_factory):
    flow = element_factory.create(UML.ControlFlow)
    item = diagram.create(ControlFlowItem, subject=flow)

    assert item.subject is flow


def test_name(create):
    """Test updating of flow name text."""
    flow = create(ControlFlowItem, UML.ControlFlow)
    name = flow.shape_tail.children[1]

    flow.subject.name = "Blah"

    assert "Blah" == name.child.text()

    flow.subject = None

    assert "" == name.child.text()


def test_guard_text_update(create):
    flow = create(ControlFlowItem, UML.ControlFlow)
    guard = flow.shape_middle

    assert "" == guard.child.text()

    specification = flow.model.create(UML.LiteralString)
    specification.value = "GuardMe"
    specification.owningConstraint = flow.subject.guard
    flow.subject.guard = specification
    assert "[GuardMe]" == guard.child.text()

    flow.subject = None
    assert "" == guard.child.text()


def test_draw(create, diagram):
    flow = create(ControlFlowItem, UML.ControlFlow)
    context = DrawContext(
        cairo=instant_cairo_context(),
        style={},
        hovered=True,
        focused=True,
        selected=True,
        dropzone=False,
    )
    diagram.update({flow})
    flow.draw(context)



from gaphor.diagram.connectors import Connector
from gaphor.diagram.copypaste import copy_full, paste_link
from gaphor.diagram.presentation import connect as _connect


def allow(line, handle, item, port=None) -> bool:
    if port is None and len(item.ports()) > 0:
        port = item.ports()[0]

    adapter = Connector(item, line)
    return adapter.allow(handle, port)


def connect(line, handle, item, port=None):
    """Connect line's handle to an item.

    If port is not provided, then first port is used.
    """
    _connect(line, handle, item)

    cinfo = line.diagram.connections.get_connection(handle)
    assert cinfo.connected is item
    assert cinfo.port


def disconnect(line, handle):
    """Disconnect line's handle."""
    diagram = line.diagram

    diagram.connections.disconnect_item(line, handle)
    assert not diagram.connections.get_connection(handle)


def get_connected(item, handle):
    assert handle in item.handles()
    if cinfo := item.diagram.connections.get_connection(handle):
        return cinfo.connected  # type: ignore[no-any-return]
    return None


def clear_model(diagram, element_factory, retain=None):
    """Clear the model and diagram, leaving only an empty diagram."""
    if retain is None:
        retain = []
    for element in list(element_factory.values()):
        if element is not diagram and element not in retain:
            element.unlink()

    for item in diagram.get_all_items():
        item.unlink()


def copy_and_paste_link(items, diagram, element_factory, retain=None):
    if retain is None:
        retain = []
    buffer = copy_full(items)
    return paste_link(buffer, diagram)


def copy_clear_and_paste_link(items, diagram, element_factory, retain=None):
    if retain is None:
        retain = []
    buffer = copy_full(items)
    clear_model(diagram, element_factory, retain)
    return paste_link(buffer, diagram)


def find(widget, name):
    if widget.get_buildable_id() == name:
        return widget
    if sibling := widget.get_next_sibling():
        if found := find(sibling, name):
            return found
    if child := widget.get_first_child():
        if found := find(child, name):
            return found

    return None



import pytest

from gaphor.core.modeling import Diagram
from gaphor.diagram.general import Box, Diamond, Ellipse, Line, MetadataItem
from gaphor.diagram.general.generalpropertypages import (
    LabelPropertyPage,
    LineEndPropertyPage,
    MetadataPropertyPage,
)
from gaphor.diagram.general.simpleitem import LineEndStyle
from gaphor.diagram.tests.fixtures import find


@pytest.mark.parametrize("item_type", [Box, Diamond, Ellipse, Line])
def test_name_page(element_factory, event_manager, item_type):
    diagram = element_factory.create(Diagram)
    item = diagram.create(item_type)
    property_page = LabelPropertyPage(item, event_manager)
    widget = property_page.construct()
    name = find(widget, "label-entry")
    name.set_text("A new label")

    assert item.label == "A new label"


def test_name_page_disposal(element_factory, event_manager):
    diagram = element_factory.create(Diagram)
    item = diagram.create(Box)
    property_page = LabelPropertyPage(item, event_manager)
    binding_ref = property_page.binding.weak_ref()
    widget_ref = property_page.construct().weak_ref()
    property_page.close()
    del property_page

    assert widget_ref() is None
    assert binding_ref() is None


def test_metadata_property_page(diagram, event_manager):
    metadata = diagram.create(MetadataItem)

    property_page = MetadataPropertyPage(metadata, event_manager)

    widget = property_page.construct()
    description = find(widget, "description")
    description.set_text("my text")

    assert metadata.description == "my text"


def test_line_end_property_page_head_and_tail(element_factory, event_manager):
    diagram = element_factory.create(Diagram)
    line = diagram.create(Line)

    property_page = LineEndPropertyPage(line, event_manager)
    widget = property_page.construct()

    head = find(widget, "line-head-end")
    tail = find(widget, "line-tail-end")

    head.set_selected(2)
    tail.set_selected(3)

    assert line.head_end == LineEndStyle.triangle
    assert line.tail_end == LineEndStyle.diamond


def test_line_end_property_page_allows_independent_values(
    element_factory, event_manager
):
    diagram = element_factory.create(Diagram)
    line = diagram.create(Line)

    property_page = LineEndPropertyPage(line, event_manager)
    widget = property_page.construct()

    head = find(widget, "line-head-end")
    tail = find(widget, "line-tail-end")

    head.set_selected(1)
    tail.set_selected(0)

    assert line.head_end == LineEndStyle.arrow
    assert line.tail_end == LineEndStyle.none



from gaphor.diagram.tests.fixtures import find
from gaphor.ui import APPLICATION_ID
from gaphor.ui.greeter import Greeter


class RecentManagerStub:
    def __init__(self, *items):
        self.items = items

    def get_items(self):
        return self.items


class RecentInfoStub:
    def __init__(self, uri):
        self.uri = uri

    def get_applications(self):
        return APPLICATION_ID

    def exists(self):
        return True

    def get_uri(self):
        return self.uri

    def get_uri_display(self):
        return self.uri


def test_greeter_with_recent_files(event_manager):
    recent_manager = RecentManagerStub(RecentInfoStub("file://path/to/foo.gaphor"))
    greeter = Greeter(None, event_manager, recent_manager)

    greeter.open()
    widget = greeter.greeter

    recent_files = find(widget, "recent-files")

    assert recent_files.get_visible()


def test_greeter_with_no_recent_files(event_manager):
    recent_manager = RecentManagerStub()
    greeter = Greeter(None, event_manager, recent_manager)

    greeter.open()
    widget = greeter.greeter
    recent_files = find(widget, "recent-files")

    assert not recent_files.get_visible()



import pytest

from gaphor.core.styling import (
    CompiledStyleSheet,
    compile_style_sheet,
)
from gaphor.core.styling.declarations import PrefersColorScheme, WhiteSpace
from gaphor.core.styling.pseudo import compute_pseudo_element_style
from gaphor.core.styling.tests.test_compiler import Node


def first_decl_block(css):
    _, value = next(compile_style_sheet(css))
    return value


def test_empty_css():
    css = ""
    rules = list(compile_style_sheet(css))

    assert not rules


def test_css_content_for_integer():
    css = """
    * {
        font-size: 42;
    }
    """
    props = first_decl_block(css)

    assert props["font-size"] == 42


def test_css_content_for_string():
    css = """
    * {
        font-family: 'sans';
    }
    """
    props = first_decl_block(css)

    assert props["font-family"] == "sans"


def test_css_content_for_ident():
    css = """
    * {
        font-family: sans;
    }
    """
    props = first_decl_block(css)

    assert props["font-family"] == "sans"


def test_css_content_for_tuple():
    css = """
    * {
        padding: 1 2 3 4;
    }
    """
    props = first_decl_block(css)

    assert props["padding"] == (1.0, 2.0, 3.0, 4.0)


def test_multiple_rules():
    css = """
    * {}
    item {}
    """

    rules = list(compile_style_sheet(css))

    assert len(rules) == 2


def test_combined_rules():
    css = """
    foo, bar {}
    """
    rules = list(compile_style_sheet(css))

    assert len(rules) == 2


def test_invalid_css_syntax():
    css = """
    foo/bar ;
    """
    rules = list(compile_style_sheet(css))
    selector, _ = rules[0]

    assert len(rules) == 1
    assert selector == "error"


def test_invalid_css_syntax_in_declarations():
    css = """
    foo { foo : bar : baz }
    """
    rules = list(compile_style_sheet(css))
    _, decls = rules[0]

    assert len(rules) == 1
    assert len(decls) == 0


def test_invalid_selector():
    css = """
    foo/bar {}
    good {}
    """
    rules = list(compile_style_sheet(css))
    selector, _ = rules[0]

    assert len(rules) == 2
    assert selector == "error"


def test_css_declaration_without_semicolon():
    css = """
    * {
        font-family: 'sans'
    }
    """
    props = first_decl_block(css)

    assert props.get("font-family") == "sans"


def test_css_multiple_declarations_without_semicolons():
    css = """
    * {
        font-family: sans;
        font-size: 42
    }
    """
    props = first_decl_block(css)

    assert props.get("font-family") == "sans"
    assert props.get("font-size") == 42


def test_compiled_style_sheet():
    css = """
    * {
        font-size: 42;
        font-family: overridden
    }
    mytype {
        font-family: sans
    }
    """

    compiled_style_sheet = CompiledStyleSheet(css)
    props = compiled_style_sheet.compute_style(Node("mytype"))

    assert props.get("font-family") == "sans"
    assert props.get("font-size") == 42


@pytest.mark.parametrize(
    "font_size", ["x-small", "small", "medium", "large", "x-large"]
)
def test_relative_font_size(font_size):
    css = f"* {{ font-size: {font_size} }}"

    props = first_decl_block(css)
    assert props.get("font-size") == font_size


@pytest.mark.parametrize("font_size", ["foo", "xx-small", "xx-large"])
def test_faulty_font_size(font_size):
    css = f"* {{ font-size: {font_size} }}"

    props = first_decl_block(css)
    assert props.get("font-size") is None


def filter_private(props):
    return {n: v for n, v in props.items() if not n.startswith("-gaphor-")}


def test_empty_compiled_style_sheet():
    css = ""

    compiled_style_sheet = CompiledStyleSheet(css)
    props = compiled_style_sheet.compute_style(Node("mytype"))

    assert filter_private(props) == {}


def test_color():
    css = "mytype { color: #00ff00 }"

    compiled_style_sheet = CompiledStyleSheet(css)
    props = compiled_style_sheet.compute_style(Node("mytype"))

    assert props.get("color") == (0, 1, 0, 1)


def test_color_typing_in_progress():
    css = "mytype { color: # }"

    compiled_style_sheet = CompiledStyleSheet(css)
    props = compiled_style_sheet.compute_style(Node("mytype"))

    assert props.get("color") is None


@pytest.mark.parametrize(
    "css_value,result",
    [
        ["normal", 0.0],
        ["sloppy", 0.5],
        ["sloppy 0.3", 0.3],
        ["sloppy -0.1", -0.1],
        ["sloppy 2.1", 2.0],
        ["sloppy -2.1", -2.0],
        ["sloppy wrong", 0.5],
    ],
)
def test_line_style(css_value, result):
    css = f"mytype {{ line-style: {css_value} }}"

    compiled_style_sheet = CompiledStyleSheet(css)
    props = compiled_style_sheet.compute_style(Node("mytype"))

    assert props.get("line-style") == result


@pytest.mark.parametrize(
    "css_value,result",
    [
        [0.5, 0.5],
        [0, 0.0],
        [1, 1.0],
        [-0.1, None],
        [1.1, None],
        ["wrong", None],
    ],
)
def test_opacity(css_value, result):
    css = f"mytype {{ opacity: {css_value} }}"

    compiled_style_sheet = CompiledStyleSheet(css)
    props = compiled_style_sheet.compute_style(Node("mytype"))

    assert props.get("opacity") == result


def test_broken_line_style():
    # diagram css is missing the closing bracket
    css = "diagram { line-style: sloppy * { }"

    compiled_style_sheet = CompiledStyleSheet(css)
    props = compiled_style_sheet.compute_style(Node("mytype"))

    assert props.get("line-style") is None


def test_variable():
    css = "diagram { --myvar: 12; line-width: var(--myvar) }"

    compiled_style_sheet = CompiledStyleSheet(css)
    props = compiled_style_sheet.compute_style(Node("diagram"))

    assert props.get("line-width") == 12


def test_variable_color():
    css = "* { --mycolor: #123456 } diagram { color: var(--mycolor) }"

    compiled_style_sheet = CompiledStyleSheet(css)
    props = compiled_style_sheet.compute_style(Node("diagram"))

    assert props.get("color") == pytest.approx(
        (0x12 / 255.0, 0x34 / 255.0, 0x56 / 255.0, 1.0)
    )


def test_unknown_variable():
    css = "diagram { line-width: var(--myvar) }"

    compiled_style_sheet = CompiledStyleSheet(css)
    props = compiled_style_sheet.compute_style(Node("diagram"))

    assert props.get("line-width") is None


def test_unknown_variable_should_use_original_value():
    css = "* { line-width: 1.0 } diagram { line-width: var(--myvar) }"

    compiled_style_sheet = CompiledStyleSheet(css)
    props = compiled_style_sheet.compute_style(Node("diagram"))

    assert props.get("line-width") == pytest.approx(1.0)


def test_unknown_variable_resolve_original_value():
    css = "* { line-width: var(--lw); --lw: 1.0 } diagram { line-width: var(--myvar) }"

    compiled_style_sheet = CompiledStyleSheet(css)
    props = compiled_style_sheet.compute_style(Node("diagram"))

    assert props.get("line-width") == pytest.approx(1.0)


def test_variable_cannot_contain_a_variable():
    css = "* { line-width: var(--a); --a: var(--b); --b: 1.0 }"

    compiled_style_sheet = CompiledStyleSheet(css)
    props = compiled_style_sheet.compute_style(Node("diagram"))

    assert props.get("line-width") is None


def test_variable_with_property():
    css = "* { line-width: var(line-width); }"

    compiled_style_sheet = CompiledStyleSheet(css)
    props = compiled_style_sheet.compute_style(Node("diagram"))

    assert props.get("line-width") is None


def test_color_schemes():
    css = """
        * { line-width: 1; }
        @media dark-mode {
         * { line-width: 2; }
        }
        @media light-mode {
         * { line-width: 3; }
        }
    """

    normal_props = CompiledStyleSheet(css).compute_style(Node("node"))
    dark_props = CompiledStyleSheet(
        css, prefers_color_scheme=PrefersColorScheme.DARK
    ).compute_style(Node("node"))
    light_props = CompiledStyleSheet(
        css, prefers_color_scheme=PrefersColorScheme.LIGHT
    ).compute_style(Node("node"))

    assert normal_props.get("line-width") == pytest.approx(1.0)
    assert dark_props.get("line-width") == pytest.approx(2.0)
    assert light_props.get("line-width") == pytest.approx(3.0)


@pytest.mark.parametrize(
    "white_space_value,result",
    [
        ["normal", WhiteSpace.NORMAL],
        ["nowrap", WhiteSpace.NOWRAP],
        ["none", None],
    ],
)
def test_white_space_normal(white_space_value, result):
    css = f"* {{ white-space: {white_space_value} }}"

    compiled_style_sheet = CompiledStyleSheet(css)
    props = compiled_style_sheet.compute_style(Node("node"))

    assert props.get("white-space") is result


def test_pseudo_element():
    css = "class::after { content: 'Hi' }"

    compiled_style_sheet = CompiledStyleSheet(css)
    props = compiled_style_sheet.compute_style(Node("class", pseudo="after"))

    assert props.get("content") == "Hi"


def test_bare_pseudo_element():
    css = "::after { content: 'Hi' }"

    compiled_style_sheet = CompiledStyleSheet(css)
    props = compiled_style_sheet.compute_style(Node("node", pseudo="after"))

    assert props.get("content") == "Hi"


def test_combined_normal_and_pseudo_element():
    css = "node, class::after { content: 'Hi' }"

    compiled_style_sheet = CompiledStyleSheet(css)
    node_props = compiled_style_sheet.compute_style(Node("node"))
    class_props = compiled_style_sheet.compute_style(Node("class", pseudo="after"))

    assert node_props.get("content") == "Hi"
    assert class_props.get("content") == "Hi"


def test_pseudo_element_inherits_normal_declarations():
    css = "node { color: blue} ::after { content: 'Hi' }"

    compiled_style_sheet = CompiledStyleSheet(css)
    props = compiled_style_sheet.compute_style(Node("node"))
    after = compute_pseudo_element_style(props, "after")

    assert after
    assert after.get("content") == "Hi"



from gaphor.core.modeling import Diagram
from gaphor.diagram.general import Line
from gaphor.diagram.propertypages import (
    InternalsPropertyPage,
    LineStylePage,
    NamePropertyPage,
    NotePropertyPage,
)
from gaphor.diagram.tests.fixtures import find
from gaphor.UML import Comment
from gaphor.UML.general import CommentItem


def test_name_page(element_factory, event_manager):
    diagram = element_factory.create(Diagram)
    property_page = NamePropertyPage(diagram, event_manager)
    widget = property_page.construct()
    name = find(widget, "name-entry")
    name.set_text("A new note")

    assert diagram.name == "A new note"


def test_line_style_page_rectilinear(diagram, event_manager):
    item = diagram.create(Line)
    property_page = LineStylePage(item, event_manager)
    widget = property_page.construct()
    line_rectangular = find(widget, "line-rectilinear")

    line_rectangular.set_active(True)

    assert item.orthogonal


def test_line_style_page_orientation(diagram, event_manager):
    item = diagram.create(Line)
    property_page = LineStylePage(item, event_manager)
    widget = property_page.construct()
    flip_orientation = find(widget, "flip-orientation")
    flip_orientation.set_active(True)

    assert item.horizontal


def test_note_page_with_item_without_subject(diagram, event_manager):
    item = diagram.create(Line)
    property_page = NotePropertyPage(item, event_manager)
    widget = property_page.construct()

    assert not widget


def test_note_page_with_item_with_subject(create, event_manager):
    item = create(CommentItem, Comment)
    property_page = NotePropertyPage(item, event_manager)
    widget = property_page.construct()

    note = find(widget, "note")
    note.get_buffer().set_text("A new note")

    assert item.subject.note == "A new note"


def test_note_page_with_subject(element_factory, event_manager):
    comment = element_factory.create(Comment)
    property_page = NotePropertyPage(comment, event_manager)
    widget = property_page.construct()

    note = find(widget, "note")
    note.get_buffer().set_text("A new note")

    assert comment.note == "A new note"


def test_internals_page_for_presentation(create):
    subject = create(CommentItem, Comment)
    property_page = InternalsPropertyPage(subject)
    widget = property_page.construct()

    internals = find(widget, "internals")
    text = internals.get_label()

    assert "CommentItem" in text
    assert "gaphor.UML.Comment" in text



"""Tests for grouping functionality in Gaphor."""

from gaphor import UML
from gaphor.diagram.group import group, ungroup
from gaphor.UML.actions import ActionItem, PartitionItem


def test_add_node_to_activity(element_factory):
    node = element_factory.create(UML.Action)
    activity = element_factory.create(UML.Activity)

    assert group(activity, node)
    assert node in activity.node
    assert node.owner is activity


def test_ungroup_node_to_activity(element_factory):
    node = element_factory.create(UML.Action)
    activity = element_factory.create(UML.Activity)

    assert group(activity, node)
    assert ungroup(activity, node)
    assert node not in activity.node
    assert not node.owner


def test_no_subpartition_when_nodes_in(element_factory):
    """Test adding subpartition when nodes added."""
    p = element_factory.create(UML.ActivityPartition)
    a1 = element_factory.create(UML.Action)
    p1 = element_factory.create(UML.ActivityPartition)
    p2 = element_factory.create(UML.ActivityPartition)

    group(p, p1)
    group(p1, a1)
    assert not group(p1, p2)


def test_action_grouping(element_factory):
    """Test adding action to partition."""
    p1 = element_factory.create(UML.ActivityPartition)
    p2 = element_factory.create(UML.ActivityPartition)
    a1 = element_factory.create(UML.Action)
    a2 = element_factory.create(UML.Action)

    group(p1, a1)
    assert len(p1.node) == 1

    group(p1, p2)
    group(p2, a1)

    assert len(p2.node) == 1
    group(p2, a2)
    assert len(p2.node) == 2


def test_ungrouping(element_factory):
    """Test action and partition removal."""
    p = element_factory.create(UML.ActivityPartition)
    a1 = element_factory.create(UML.Action)
    a2 = element_factory.create(UML.Action)

    group(p, a1)
    group(p, a2)

    ungroup(p, a1)
    assert len(p.node) == 1
    ungroup(p, a2)
    assert len(p.node) == 0


def test_ungrouping_with_actions(create):
    """Test partition with actions removal."""
    p = create(PartitionItem, UML.ActivityPartition)
    a1 = create(ActionItem, UML.Action)
    a2 = create(ActionItem, UML.Action)
    partition = p.subject

    group(partition, a1.subject)
    group(partition, a2.subject)

    assert len(partition.node) == 2, partition.node

    ungroup(p.subject, a1.subject)
    ungroup(p.subject, a2.subject)

    assert len(p.children) == 0
    assert len(partition.node) == 0



import pytest

from gaphor.storage.recovery import EventLog, sha256sum


@pytest.fixture
def test_file(tmp_path):
    f = tmp_path / "testfile"
    f.write_bytes(b"abc")
    return f


@pytest.fixture
def event_log(test_file):
    event_log = EventLog("_", test_file)
    return event_log


def test_sha256sum(tmp_path):
    tmp_file = tmp_path / "testfile"
    with tmp_file.open(mode="wb") as f:
        f.write(b"abcdefg")

    assert (
        sha256sum(tmp_file)
        == "7d1a54127b222502f5b79b5fb0803061152a44f92b37e23c6527baf665d4da9a"
    )


def test_read_event_log(event_log):
    event_log.write(["my", "line"])

    lines = list(event_log.read())

    assert ["my", "line"] in lines


def test_should_not_read_if_file_changed(event_log, test_file):
    event_log.write(["my", "line"])
    test_file.write_bytes(b"123")

    lines = list(event_log.read())

    assert ["my", "line"] not in lines


def test_clear_event_log(event_log):
    event_log.write(["my", "line"])

    event_log.clear()

    lines = list(event_log.read())

    assert not lines


def test_clear_and_add_event_log(event_log):
    event_log.write(["my", "line"])
    event_log.clear()
    event_log.write(["another line"])

    lines = list(event_log.read())

    assert ["my", "line"] not in lines
    assert ["another line"] in lines


def test_move_aside_event_log(event_log):
    event_log.write(["my", "line"])

    event_log.move_aside()

    assert not event_log.log_file.exists()
    assert event_log.log_file.with_suffix(".recovery.bak").exists()



import pytest
from gaphas.item import NW, SE

from gaphor import UML
from gaphor.diagram.drop import grow_parent
from gaphor.diagram.tools.dropzone import drop_zone_tool, on_motion
from gaphor.UML.deployments import ArtifactItem, NodeItem


@pytest.mark.asyncio
async def test_hover_over_drop_zone(diagram, element_factory, view):
    node_item = diagram.create(NodeItem, subject=element_factory.create(UML.Node))
    node_item.width = node_item.height = 200
    comp_item = diagram.create(
        ArtifactItem, subject=element_factory.create(UML.Artifact)
    )

    view.request_update((node_item, comp_item))
    await view.update()

    tool = drop_zone_tool(ArtifactItem, UML.Artifact)
    view.add_controller(tool)
    on_motion(tool, 100, 100, ArtifactItem, UML.Artifact)

    selection = view.selection
    assert selection.dropzone_item is node_item


def test_grow_parent(diagram, element_factory):
    node_item = diagram.create(NodeItem, subject=element_factory.create(UML.Node))
    node_item.width = node_item.height = 200
    comp_item = diagram.create(
        ArtifactItem, subject=element_factory.create(UML.Artifact)
    )
    comp_item.matrix.translate(200, 200)

    grow_parent(node_item, comp_item)

    assert node_item.handles()[NW].pos.tuple() == (0, 0)
    assert node_item.handles()[SE].pos.tuple() == (320, 270)



"""For all relevant model elements, check if there is at least one "owner"
("owner" is a derived union).

This is needed to display all elements in the tree view.

NOTE: Comment does not have an owner. In the model it does, though.
"""

import itertools

import pytest

import gaphor.SysML.diagramitems
import gaphor.UML.diagramitems
from gaphor import UML
from gaphor.core.modeling.properties import derived
from gaphor.diagram.support import get_model_element


def all_subset_properties(prop):
    for sub in prop.subsets:
        if isinstance(sub, derived):
            yield from all_subset_properties(sub)
        else:
            yield sub


def all_presented_elements(module):
    return (
        get_model_element(getattr(module, name))
        for name in dir(module)
        if not name.startswith("_")
        and get_model_element(getattr(module, name) not in (None, UML.Comment))
    )


def all_presented_uml_and_sysml_elements():
    return itertools.chain(
        all_presented_elements(gaphor.UML.diagramitems),
        all_presented_elements(gaphor.SysML.diagramitems),
        [
            UML.ExecutionOccurrenceSpecification,
            UML.ExtensionEnd,
            UML.MessageOccurrenceSpecification,
        ],
    )


def concrete_owner_property(class_: UML.Element):
    return (
        p
        for p in class_.__properties__
        if p in all_subset_properties(UML.Element.owner)
    )


def test_all_presented_uml_and_sysml_elements():
    elements = all_presented_uml_and_sysml_elements()
    assert all(issubclass(c, UML.Element) for c in elements)


@pytest.mark.parametrize("class_", all_presented_uml_and_sysml_elements())
def test_element_has_concrete_ownable_property(class_):
    owners = list(concrete_owner_property(class_))

    assert any(owners)



"""Basic test case for Gaphor tests.

Everything is about services so the Case can define it's required
services and start off.
"""

# ruff: noqa: I001
from __future__ import annotations

from io import StringIO
from pathlib import Path

import pytest
import pytest_asyncio


# Load gaphor.ui first, so GTK library versions are set corrently
import gaphor.ui  # noqa: F401


from gaphor.core import Transaction
from gaphor.core.eventmanager import EventManager
from gaphor.core.modeling import ElementFactory
from gaphor.core.modeling.elementdispatcher import ElementDispatcher
from gaphor.core.modeling.modelinglanguage import (
    CoreModelingLanguage,
    MockModelingLanguage,
)
from gaphor.diagram.general.modelinglanguage import GeneralModelingLanguage
import gaphor.storage as storage
from gaphor.SysML.modelinglanguage import SysMLModelingLanguage
from gaphor.UML.modelinglanguage import UMLModelingLanguage
from gaphor.UML.sanitizerservice import SanitizerService
from gaphor.UML.uml import Diagram
from gaphor.ui.diagramview import DiagramView


@pytest.fixture(autouse=True)
def assert_not_in_transaction():
    assert not Transaction.in_transaction()


@pytest.fixture
def event_manager():
    return EventManager()


@pytest.fixture
def element_factory(event_manager, modeling_language):
    element_factory = ElementFactory(
        event_manager, ElementDispatcher(event_manager, modeling_language)
    )
    yield element_factory
    element_factory.shutdown()


@pytest.fixture
def modeling_language():
    return MockModelingLanguage(
        CoreModelingLanguage(),
        GeneralModelingLanguage(),
        UMLModelingLanguage(),
        SysMLModelingLanguage(),
    )


@pytest.fixture
def sanitizer_service(event_manager):
    sanitizer_service = SanitizerService(event_manager)
    yield sanitizer_service
    sanitizer_service.shutdown()


@pytest.fixture
def diagram(element_factory, event_manager):
    with Transaction(event_manager):
        diagram = element_factory.create(Diagram)
    yield diagram
    with Transaction(event_manager):
        diagram.unlink()


@pytest.fixture
def create(diagram, element_factory):
    def _create(item_class, element_class=None):
        return diagram.create(
            item_class,
            subject=(element_factory.create(element_class) if element_class else None),
        )

    return _create


@pytest.fixture
def saver(element_factory):
    def save():
        """Save diagram into string."""

        f = StringIO()
        storage.save(f, element_factory)
        data = f.getvalue()
        f.close()

        return data

    return save


@pytest.fixture
def loader(element_factory, modeling_language):
    def load(data):
        """Load data from specified string."""
        element_factory.flush()
        assert not list(element_factory.select())

        f = StringIO(data)
        storage.load(
            f, element_factory=element_factory, modeling_language=modeling_language
        )
        f.close()

    return load


@pytest.fixture
def test_models():
    """The folder where test models can be found."""
    return Path(__file__).parent.parent / "test-models"


@pytest.fixture
def models():
    """The folder where test models can be found."""
    return Path(__file__).parent.parent / "models"


@pytest_asyncio.fixture
async def view(diagram):
    view = DiagramView(model=diagram)
    yield view
    diagram.unregister_view(view)


@pytest.fixture(autouse=True)
def tmp_get_cache_config_dir(tmp_path, monkeypatch):
    monkeypatch.setattr(gaphor.settings, "get_config_dir", lambda: tmp_path)
    monkeypatch.setattr(gaphor.settings, "get_cache_dir", lambda: tmp_path)



import pytest

from gaphor import UML
from gaphor.core.modeling import StyleSheet
from gaphor.UML.states.drop import drop_region
from gaphor.UML.states.state import StateItem


@pytest.fixture
def parent_state(create, diagram, element_factory, view):
    element_factory.create(StyleSheet)
    parent: StateItem = create(StateItem, UML.State)
    parent.subject.region = element_factory.create(UML.Region)
    parent.subject.region = element_factory.create(UML.Region)
    diagram.update()
    view.update_bounding_box(diagram.ownedPresentation)
    return parent


def test_drop_state_on_first_region(parent_state, create):
    state: StateItem = create(StateItem, UML.State)

    drop_region(state, parent_state, 10, parent_state.height / 2)

    assert state.subject.container is parent_state.subject.region[0]


def test_drop_state_on_second_region(parent_state, create):
    state: StateItem = create(StateItem, UML.State)

    drop_region(state, parent_state, 10, parent_state.height - 10)

    assert state.subject.container is parent_state.subject.region[1]


def test_drop_from_first_region_on_second_region(parent_state, create):
    state: StateItem = create(StateItem, UML.State)

    drop_region(state, parent_state, 10, parent_state.height / 2)
    drop_region(state, parent_state, 10, parent_state.height - 10)

    assert state.subject.container is parent_state.subject.region[1]



import pytest

from gaphor import UML
from gaphor.ui.clipboard import Clipboard
from gaphor.UML.classes import PackageItem
from gaphor.UML.general import CommentItem


class MockSystemClipboard:
    def __init__(self):
        self.copy_buffer = None

    def set_content(self, content_provider):
        self.copy_buffer = content_provider.get_value()

    async def read_value_async(self, gtype, io_priority):
        return self.copy_buffer


@pytest.fixture
def clipboard(event_manager, element_factory, monkeypatch):
    monkeypatch.setattr("gi.repository.Gdk.ContentProvider.new_union", lambda v: v[0])
    monkeypatch.setattr("gi.repository.Gdk.ContentProvider.new_for_value", lambda v: v)

    return Clipboard(event_manager, element_factory, MockSystemClipboard())


@pytest.mark.asyncio
async def test_copy_link(clipboard, diagram, view, element_factory):
    ci = diagram.create(CommentItem, subject=element_factory.create(UML.Comment))
    view.selection.select_items(ci)

    clipboard.copy(view)
    assert list(diagram.get_all_items()) == [ci]

    clipboard.paste_link(view)
    await clipboard.gather_background_task()

    assert ci in diagram.get_all_items()
    assert len(list(diagram.get_all_items())) == 2, list(diagram.get_all_items())


@pytest.mark.asyncio
async def test_copy_full_with_owned_element(clipboard, diagram, view, element_factory):
    package_item = diagram.create(
        PackageItem, subject=element_factory.create(UML.Package)
    )
    package_item.subject.nestedPackage = element_factory.create(UML.Package)
    view.selection.select_items(package_item)

    clipboard.copy(view)

    clipboard.paste_full(view)
    await clipboard.gather_background_task()

    assert len(element_factory.lselect(UML.Package)) == 4
